/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * ***** END LICENSE BLOCK ***** */



/**
 * glActionBuffer provides an abstraction layer for running actions on mail
 * and having them be transparently buffered and run in the background.
 *
 * @class ?
 */
var glActionBuffer = glBase.extend({
    // Properties {
    /**
     * Whether to auto flush the action buffer or not, it defaults to true
     * @property {Boolean} ?
     */
    autoflush: true,

    /**
     * glGmail object
     * @property {glGmail} ?
     */
    _glGmail: null,

    /**
     * The buffered actions in a hash by action type
     * @property {Object} ?
     */
     _actions: null,

    /**
     * References to all currently active glRequest objects.
     * Used to ensure that they are garbage collected when the request is completed.
     * @property {Array} ?
     */
     _requests: null,
    //}

    /**
     * The constructor.
     * @constructor ?
     * @param {glGmail} gmail - The glGmail object for the page
     */
    constructor: function(gmail) {
        this._glGmail = gmail;
        this._actions = {
            addLabel: {},
            removeLabel: {},
            addStar: [],
            removeStar: [],
            deleteThread: [],
            spamThread: [],
            notSpamThread: [],
            markThreadRead: [],
        };
        this._requests = [];
    },

    // Methods {
    // Action Methods {
    /**
     * Archives the given thread. Alias for <code>removeLabelFromThread(threadId,
     * '^i')</code>.
     * @function ?
     * @param {String} threadId - The id of the thread
     */
    archiveThread: function(threadId) {
        this.removeLabelFromThread(threadId, '^i');
    },

    /**
     * Adds the given label to the thread
     * @function ?
     * @param {String} threadId - The id of the thread
     * @param {String} label - The label to add
     */
    addLabelToThread: function(threadId, label) {
        if(!this._actions.addLabel[label]) {
            this._actions.addLabel[label] = [threadId];
        } else {
            this._actions.addLabel[label].push(threadId);
        }

        if(this.autoflush) this.flush();
    },

    /**
     * Removes the given label from the thread
     * @function ?
     * @param {String} threadId - The id of the thread
     * @param {String} label - The label to remove
     */
    removeLabelFromThread: function(threadId, label) {
        if(!this._actions.removeLabel[label]) {
            this._actions.removeLabel[label] = [threadId];
        } else {
            this._actions.removeLabel[label].push(threadId);
        }

        if(this.autoflush) this.flush();
    },

    /**
     * Stars the thread
     * @function ?
     * @param {String} threadId - The id of the thread
     */
    addStarToThread: function(threadId) {
        this._actions.addStar.push(threadId);

        if(this.autoflush) this.flush();
    },

    /**
     * Unstars the thread
     * @function ?
     * @param {String} threadId - The id of the thread
     */
    removeStarFromThread: function(threadId) {
        this._actions.removeStar.push(threadId);

        if(this.autoflush) this.flush();
    },

    /**
     * Deletes the given thread
     * @function ?
     * @param {String} threadId - The id of the thread
     */
    deleteThread: function(threadId) {
        this._actions.deleteThread.push(threadId);

        if(this.autoflush) this.flush();
    },

    /**
     * Spams the given thread
     * @function ?
     * @param {String} threadId - The id of the thread
     */
    spamThread: function(threadId) {
        this._actions.spamThread.push(threadId);

        if(this.autoflush) this.flush();
    },

    /**
     * Unspams the given thread
     * @function ?
     * @param {String} threadId - The id of the thread
     */
    notSpamThread: function(threadId) {
        this._actions.notSpamThread.push(threadId);

        if(this.autoflush) this.flush();
    },
    
    /**
    * Mark a thread as read
    * @function ?
    * @param {String} threadId - The id of the thread
    */
    markThreadRead: function(threadId){
    	this._actions.markThreadRead.push(threadId);
    	if(this.autoflush) this.flush();
    },
    
    //}

    // Buffer Control Methods {
    /**
     * Called on shutdown, it flushs synchronously to make sure changes get saved
     * @function ?
     */
    shutdown: function() {
        this.flush(false);
    },

    /**
     * Flushes the buffered actions, running them all
     * @function ?
     * @param {Boolean} async - Whether or not to do asynchronously - defaults to true
     */
    flush: function(async) {
        if(typeof async == 'undefined') async = true;

        var requests = [];

        // Process addLabel actions
        for(var label in this._actions.addLabel) {
            if(this._actions.addLabel[label].length) {
                requests.push(glRequest.createAddLabelRequest(this._glGmail, label, this._actions.addLabel[label]));
                delete this._actions.addLabel[label];
            }
        }

        // Process removeLabel actions
        for(var label in this._actions.removeLabel) {
            if(this._actions.removeLabel[label].length) {
                requests.push(glRequest.createRemoveLabelRequest(this._glGmail, label, this._actions.removeLabel[label]));
                delete this._actions.removeLabel[label];
            }
        }

        // Process addStar actions
        if(this._actions.addStar.length) {
            requests.push(glRequest.createAddStarRequest(this._glGmail, this._actions.addStar));
            this._actions.addStar = [];
        }

        // Process removeStar actions
        if(this._actions.removeStar.length) {
            requests.push(glRequest.createRemoveStarRequest(this._glGmail, this._actions.removeStar));
            this._actions.removeStar = [];
        }

        // Proccess deleteThread actions
        if(this._actions.deleteThread.length) {
            requests.push(glRequest.createDeleteThreadRequest(this._glGmail, this._actions.deleteThread));
            this._actions.deleteThread = [];
        }
        
        // Process spamThread actions
        if(this._actions.spamThread.length) {
            requests.push(glRequest.createSpamThreadRequest(this._glGmail, this._actions.spamThread));
            this._actions.spamThread = [];
        }
        
        // Process notSpamThread actions
        if(this._actions.notSpamThread.length) {
            requests.push(glRequest.createNotSpamThreadRequest(this._glGmail, this._actions.notSpamThread));
            this._actions.notSpamThread = [];
        }
        
        
        // Proccess markThreadRead actions
        if(this._actions.markThreadRead.length) {
            requests.push(glRequest.createMarkThreadRequest(this._glGmail, this._actions.markThreadRead));
            this._actions.markThreadRead = [];
        }

        // Run queued requests
        for(var i = 0; i < requests.length; i++) {
            glLogger.log('Making '+(async?'asynchronous ':'')+'request to: "'+requests[i].url+'"', 'glActionBuffer');
            this._requests.push(requests[i]);
            requests[i].addEventListener(glRequest.COMPLETED, this._requestCompletedCallback, this);
            requests[i].run(async);
        }
    },

    /**
     * Removes the request object reference from the _requests array
     * @function ?
     * @param {Event} evt - The request completed event
     */
    _requestCompletedCallback: function(evt) {
        for(var i = 0; i < this._requests.length; i++) {
            if(evt.target === this._requests[i]) {
                // Found the request in the _requests array so clean up
                glLogger.log('Request to "'+this._requests[i].url+'" completed', 'glActionBuffer');
                this._requests[i].removeEventListener(glRequest.COMPLETED, this._requestCompletedCallback);
                this._requests.splice(i, 1);
                break;
            }
        }
    },
    //}
    //}
});
