/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Manages running of actions and notifications of actions that have occurred.
 *
 * @class ?
 */
var glActionManager = glBase.extend({
    // Properties {
    /**
     * The {@link glActionBuffer} object that glActionManager uses as one of its backends
     * @property {glActionBuffer} ?
     */
    actionBuffer: null,

    /**
     * The {@link glMenuActions} object that glActionManager uses as one of its backends
     * @property {glMenuActions} ?
     */
	menuActions: null,
	
	/** @property {Object[]} ? Recently (code) requested actions */
	_recentActions: null,

    /**
     * glGmail object
     * @property {glGmail} ?
     */
    _glGmail: null,
    //}

    // Methods {
    /**
     * Connects the action manager to the given glGmail object's page.
     * @function ?
     * @param {glGmail} gmail - The glGmail object for the page
     */
    connect: function(gmail) {
        this._glGmail = gmail;
        this.actionBuffer = new glActionBuffer(gmail);
        this._recentActions = [];
		this.menuActions = new glMenuActions(gmail);
		
		gmail.doc.defaultView.wrappedJSObject.top.menuActions = this.menuActions;

        gmail.environment.addEventListener(glEnvironment.ACTION_PERFORMED, this._environmentActionPerformed, this);
    },

    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    shutdown: function() {
        this._glGmail = null;
        this.actionBuffer.shutdown();

        this._removeAllListeners();
    },

    /**
     * Performs the given action
     * @function ?
     * @paramset Action
     * @param {Object} action - The action constant for the action you want to perform
     * @param {glThread or String or Array} thread - The thread(s) you want to perform the action on
	 * @param {String} serverActionType - How to interact with server
     * @paramset Label
     * @param {Object} action - The action constant for the action you want to perform
     * @param {glThread or String or Array} thread - The thread(s) you want to perform the action on
	 * @param {String} serverActionType - How to interact with server
     * @param {String} label - The label to add or remove
     */
    performAction: function() {
        if(arguments.length < 2) throw new Error('Need at least 2 arguments');

        var action = null, threads = null, label = null, serverActionType = null;
            action = arguments[0];
            threads = arguments[1];
			serverActionType = arguments[2] || glActionManager.USE_AJAX;
            label = arguments[3];

        // If threads is not an array, make it one
        if(typeof threads != 'array') threads = [threads];


		
        for(var i = 0; i < threads.length; i++) {
            // Perform action
			var threadId = (threads[i] instanceof glThread ? threads[i].threadId : threads[i]);
			if( this.actionBuffer[action['bufferFunc']] && (serverActionType==glActionManager.USE_AJAX || !this.menuActions[action['bufferFunc']]) ){
				var args = [threadId];
            	if(label) args.push(label);
            	this.actionBuffer[action['bufferFunc']].apply(this.actionBuffer, args);
			}else if( serverActionType==glActionManager.USE_DOM ){
				var args = [];
				if( label ) args.push(label);
				this.menuActions[action['bufferFunc']].apply(this.menuActions, args);
			}

            // Notify listeners that the action was performed
            var performedAction = {action: action, itemType: 'thread', item:threadId, serverActionType: serverActionType};
            if(label) performedAction.label = label;
			if( threads[i] instanceof glThread ) performedAction.threadObject = threads[i];
            this._processAction(performedAction);
            
            if( serverActionType==glActionManager.USE_AJAX ){
            	this._glGmail.environment.requestRefresh();
            }
        }
    },

    /**
     * Adds a listener function for actions on the given thread
     * @function ?
     * @param {String} thread - The thread to listen for actions on
     * @param {Function} listener - The function to call on thread action
     * @param {optional Object} thisObj - The optional object to use as "this" when the listener is called
     */
    addThreadListener: function(thread, listener, thisObj) {
        // Use event dispatcher mixin to store and manage events because it's easier
        this.addEventListener('thread_action_'+thread, listener, thisObj);
    },

    /**
     * Removes the given thread listener
     * @function ?
     * @param {String} thread - The thread that the listener is registered for
     * @param {Function} listener - The function that was registered as a listener
     */
    removeThreadListener: function(thread, listener) {
        // Use event dispatcher mixin to store and manage events because it's easier
        this.removeEventListener('thread_action_'+thread, listener);
    },

    /**
     * Callback for the {@link glEnvironment.ACTION_PERFORMED} event. Notifies
     * listeners what event occurred.
     * @function ?
     * @param {Event} evt - The ACTION_PERFORMED event
     * @return void
     */
    _environmentActionPerformed: function(evt) {
        glLogger.log('Notified by environment that an action was performed', 'glActionManager');
        var actions = this._parseURLAndPost(evt.url, evt.postData);
        for(var i = 0; i < actions.length; i++) {
		actions[i].environmentGenerated = true;
            this._processAction(actions[i]);
        }
    },

    /**
     * Takes the given action and affected threads and notifies all listeners
     * about what happened
     * @function ?
     * @param {Object} action - The action data
     * @... {Object} action - The action constant for the performed action
     * @... {String} label - The label that was added or removed if needed
     * @... {String} item - The affected item's id
     * @... {String} itemType - Either 'thread' or 'message'
     */
    _processAction: function(action) {
        glLogger.log('Notifying listeners that an action took place', 'glActionManager');
        if(action.itemType == 'thread') {
            // Notify listeners for this thread
            var evt = {name: 'thread_action_'+action.item, action: action.action, serverActionType: action.serverActionType, threadObject: action.threadObject, environmentGenerated: action.environmentGenerated, performActionRequest: action.performActionRequest};
            if(action.label) evt.label = action.label;
            this.dispatchEvent(evt);

            // Dispatch THREAD_CHANGED event
            var evt = {name: glActionManager.THREAD_CHANGED, action: action.action, thread: action.item, serverActionType: action.serverActionType, threadObject: action.threadObject, environmentGenerated: action.environmentGenerated, performActionRequest: action.performActionRequest};
            if(action.label) evt.label = action.label;
            this.dispatchEvent(evt);

        } else {
            // Dispatch MESSAGE_CHANGED event
            var evt = {name: glActionManager.MESSAGE_CHANGED, action: action.action, message: action.item, serverActionType: action.serverActionType, threadObject: action.threadObject, environmentGenerated: action.environmentGenerated, performActionRequest: action.performActionRequest};
            if(action.label) evt.label = action.label;
            this.dispatchEvent(evt);
        }

        // Dispatch generic action event
        var evt = {name: action.action.eventName, item: action.item, itemType: action.itemType, serverActionType: action.serverActionType, threadObject: action.threadObject, environmentGenerated: action.environmentGenerated, performActionRequest: action.performActionRequest};
        if(action.label) evt.label = action.label;
        this.dispatchEvent(evt);
    },

    /**
     * Parses the given request URL and post data to determine the action and affected threads
     * @function ?
     * @param {String} url - The action request URL
     * @param {String} postData - The request post data
     * @return An array of objects that can be passed to _processAction for event dispatching
     */
    _parseURLAndPost: function(url, postData) {
        var match = /[?&]act=([^&]+)/.exec(url);
        if(!match) return [];

        glLogger.log('Action code is: '+match[1], 'glActionManager');

        var actions = [];
        var urlAction = match[1];

        // Extract affected items from post data
        postData = postData.replace(/^&+/, '').replace(/&+$/, ''); // Strip &s from beginning and end
        if(/&?t=/.test(postData)) {
            // Performed action on a thread
            var itemType = 'thread';
            var items = postData.split(/&?t=/);
        } else {
            // Performed action on message
            var itemType = 'message';
            var items = postData.split(/&?m=/);
        }
        items.shift(); // Remove empty string from begining

        // Create simple helper function that merges in the affected item and type into the given action
        var addActionForItems = function(data) {
            for(var i = 0; i < items.length; i++) {
                var action = {itemType: itemType, item: items[i]};
                action.action = data.action;
                if(data['label']) action.label = data.label;
                actions.push(action);
            }
        };

        // Determine action based on action code and process
        if(urlAction == 'arl') { // Multiple label modify
            // Calculate label actions - adds and removes
            var actionsTemp = [];
            var addLabelRegex = /[&?]acn=([^&]+)/g, removeLabelRegex = /[&?]rcn=([^&]+)/g, matches = null;
			url = decodeURIComponent(url);
            while((matches = addLabelRegex.exec(url)) != null) {
                actionsTemp.push({ action: glActionManager.ADD_LABEL, label: matches[1] });
            }
            while((matches = removeLabelRegex.exec(url)) != null) {
                actionsTemp.push({ action: glActionManager.REMOVE_LABEL, label: matches[1] });
            }

            // Go through actions and add
            for(var i = 0; i < actionsTemp.length; i++) {
                addActionForItems(actionsTemp[i]);
            }
        } else if(/^ac_/.test(urlAction)) { // Add label
            var label = decodeURIComponent(/^ac_(.*)/.exec(urlAction)[1]);
            addActionForItems({ action: glActionManager.ADD_LABEL, label: label});
        } else if(/^rc_/.test(urlAction)) { // Remove label
            var label = decodeURIComponent(/^rc_(.*)/.exec(urlAction)[1]);
            if(label == '^i') {
                addActionForItems({ action: glActionManager.ARCHIVE });
            } else {
                addActionForItems({ action: glActionManager.REMOVE_LABEL, label: label });
            }
        } else if(urlAction == 'sp') { // Report spam
            addActionForItems({ action: glActionManager.REPORT_SPAM });
        } else if(urlAction == 'us') { // UnReport spam
            addActionForItems({ action: glActionManager.NOT_SPAM });
        } else if(urlAction == 'st') { // Add star
            addActionForItems({ action: glActionManager.STAR });
        } else if(urlAction == 'xst') { // Remove star
            addActionForItems({ action: glActionManager.UNSTAR });
        } else if(urlAction == 'tr') { // Item deleted
            addActionForItems({ action: glActionManager.DELETE });
        }

        return actions;
    },
    //}
},{
    // Static Properties {
    /** @property {static Object} ? Action constant for adding a label */
    ADD_LABEL: { bufferFunc: 'addLabelToThread', eventName: 'label_added' },
    /** @property {static Object} ? Action constant for removing a label */
    REMOVE_LABEL: { bufferFunc: 'removeLabelFromThread', eventName: 'label_removed' },
    /** @property {static Object} ? Action constant for staring */
    STAR: { bufferFunc: 'addStarToThread', eventName: 'star_added' },
    /** @property {static Object} ? Action constant for unstaring */
    UNSTAR: { bufferFunc: 'removeStarFromThread', eventName: 'star_removed' },
    /** @property {static Object} ? Action constant for archiving */
    ARCHIVE: { bufferFunc: 'archiveThread', eventName: 'item_archived' },
    /** @property {static Object} ? Action constant for deleting */
    DELETE: { bufferFunc: 'deleteThread', eventName: 'item_deleted' },
    /** @property {static Object} ? Action constant for reporting spam */
    REPORT_SPAM: { bufferFunc: 'spamThread', eventName: 'item_reported_spam' },
    /** @property {static Object} ? Action constant for reporting spam */
    NOT_SPAM: { bufferFunc: 'notSpamThread', eventName: 'item_not_spam' },
    /** @property {static Object} ? Action constant for deleting */
    MARK_AS_READ: { bufferFunc: 'markThreadRead', eventName: 'item_marked_read' },

    /**
     * Dispatched when a thread is changed. Includes the action, which is a reference
     * to one of the action constants. i.e. <code>glActionManager.REMOVE_LABEL</code>.
     * It also includes the label if relevant and the thread id.
     * @event ?
     */
    THREAD_CHANGED: 'thread_changed',

    /**
     * Dispatched when a message is changed. Includes the action like {@link THREAD_CHANGED},
     * the message id, and the label if relevant.
     * @event ?
     */
    MESSAGE_CHANGED: 'message_changed',

	/** @property {static String} ? Tells performAction to use ajax if possible */
	USE_AJAX: 'ajax',
	/** @property {static String} ? Tells performAction to use dom menu clicking if possible */
	USE_DOM: 'dom',
    //}
});
glActionManager.implement(glEventDispatcher); // Add event dispatching support