/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Contains the functionality to securely interact with the underlying gmail page
 * by means of a hidden added element and events.
 *
 * @class ?
 */
var glPageWrapper = glBase.extend({
    // Properties {
    /**
     * The glGmail object for the page
     * @property {glGmail} ?
     */
    _glGmail: null,

    /**
     * Whether or not the UI is ready
     * @property {Boolean} ?
     */
    _uiIsReady: false,

    /**
     * The injected element that is used for communication
     * @property {Element} ?
     */
    _injectedElem: null,

    /**
     * The sandbox object for the page
     * @property {Sandbox} ?
     */
    _sandbox: null,
    //}

    // Methods {
    /**
     * Connects the element to the given glGmail object's page.
     * @function ?
     * @param {glGmail} gmail - The glGmail object for the page
     */
    connect: function(gmail) {
        glLogger.log('Page wrapper connected', 'glPageWrapper');
        this._glGmail = gmail;

        // Load injected code
        this._injectCode(this._loadBaseInjectedCode());

        // Get injected element and attach event listeners
        var me = this;
        this._injectedElem = this._glGmail.doc.getElementById('_glGmailInjected');
        this._injectedListener = function(e) { me._eventReceived(e); }; // So we can remove it later
        this._injectedElem.addEventListener('glGmailEventDispatched', this._injectedListener, false, true);
    },

    /**
     * Disconnects the page wrapper object from the page, removing event listeners
     * and the injected element.
     * @function ?
     */
    disconnect: function() {
        glLogger.log('Page wrapper disconnected', 'glPageWrapper');
        this._injectedElem.removeEventListener('glGmailEventDispatched', this._injectedListener, true);
        this._injectedElem.parentNode.removeChild(this._injectedElem);
        this._injectedElem = null;
        this._injectedListener = null;
        this._glGmail = null;
    },

    /**
     * Returns the object at the given offset in the page GLOBALS object.
     * @function {Object} ?
     * @param {int} offset - The offset in the GLOBALS object
     * @return The value of the offset in GLOBALS
     */
    getGlobal: function(offset) {
        return this._runInPage('getGlobal', offset);
    },

    /**
     * Returns the VIEW_DATA object for the page
     * @function {Object} ?
     * @return The VIEW_DATA object
     */
    getViewData: function() {
        return this._runInPage('getViewData');
    },

    /**
     * Returns the current view type
     * @function {String} ?
     * @return The current view type as given by the gmonkey object ("tl" =>
     * theadlist, "cv" => conversation, "co" => compose, "ct" => contacts, or
     * "s" => settings)
     */
    getActiveViewType: function() {
        return this._runInPage('getActiveViewType');
    },

    /**
     * Returns the root node of the view. The class that it is looking for is
     * added by the injected code on view change.
     * @function {Element} ?
     * @return The root node of the view
     */
    getViewRoot: function() {
        var elems = this._glGmail.canvasDocument.getElementsByClassName('glViewRoot');
        if(!elems.length) {
            return null;
        } else {
            return elems[0];
        }
    },

    /**
     * Returns whether or not the UI is ready and all of the css and javascript
     * have loaded.
     * @function {Boolean} ?
     * @return Whether the UI is ready
     */
    isUIReady: function() {
        return this._uiIsReady;
    },

    /**
     * Runs the given function in the page using a glGmailRunFunction event and
     * returns the return value of the function.
     * @function {Object} ?
     * @param {String} func - The function to call
     * @param {Array} ...rest - The arguments to the function call
     * @return The value of the evaluated code
     */
    _runInPage: function(func) {
        glLogger.log('Running function "'+func+'"', 'glPageWrapper');

        // Set attribute for code
        this._injectedElem.setAttribute('func', func);
        this._injectedElem.setAttribute('params', arguments.length-1);
        for(var i = 1; i < arguments.length; i++) {
            this._injectedElem.setAttribute('param'+i, arguments[i]);
        }

        // Dispatch event
        var evt = this._glGmail.doc.createEvent('Events');
        evt.initEvent('glGmailRunFunction', true, false);
        this._injectedElem.dispatchEvent(evt);

        // Extract return value
        var returnValue = null;
        if(this._injectedElem.hasAttribute('return')) {
            // Return value is wrapped in object and placed in data property because native JSON can only handle objects and arrays - not bare strings and numbers
            var returnValueString = this._injectedElem.getAttribute('return');
            var json = Components.classes["@mozilla.org/dom/json;1"].createInstance(Components.interfaces.nsIJSON);
            returnValue = json.decode(returnValueString).data;
        }

        // Clean up attributes
        this._injectedElem.removeAttribute('func');
        this._injectedElem.removeAttribute('return');
        this._injectedElem.removeAttribute('params');
        for(var i = 1; i < arguments.length; i++) {
            this._injectedElem.removeAttribute('param'+i);
        }

        return returnValue;
    },

    /**
     * Callback for the glGmailEventDispatched event which is dispatched by the
     * injected code. The event name is turned into a real event and re-dispatched.
     * @function ?
     * @param {Event} evt - The glGmailEventDispatched event
     */
    _eventReceived: function(evt) {
        // Get event name
        var eventName = this._injectedElem.getAttribute('event');

        glLogger.log('Dispatching event "'+eventName+'"', 'glPageWrapper');
        var event = {name: eventName};
        switch(eventName) {
            case 'ui_ready':
                this._uiIsReady = true;
                break;
            case 'view_changed':
                event.view = this.getActiveViewType();
                break;
        }
        this.dispatchEvent(event);
    },

    /**
     * Injects the given code into the page. Can be made public if a plugin
     * requests it, but it shouldn't be necessary in most cases.
     * @function ?
     * @param {String} code - The code to inject into the page
     */
    _injectCode: function(code) {
        // Create sandbox if not defined
        if(!this._sandbox) {
            this._sandbox = new Components.utils.Sandbox(this._glGmail.window);
            this._sandbox.window = this._glGmail.window;
            this._sandbox.document = this._glGmail.doc;
            this._sandbox.unsafeWindow = this._glGmail.window.wrappedJSObject;
            this._sandbox.log = glLogger.log;
            this._sandbox.__proto__ = this._glGmail.window;
        }

        // Wrap code in window scope
        var scriptToEval = "with (window?window:null) { " + code + " \n};";
        try {
            //glLogger.log("Injecting code:\n"+scriptToEval, 'glPageWrapper');
            Components.utils.evalInSandbox(scriptToEval, this._sandbox);
        } catch(e) {
            glLogger.error('Error injecting code:'+"\n"+e, 'glPageWrapper');
        }
    },

    /**
     * Loads {@link content/js/gmail/injected-code.js injected-code.js} and
     * returns the contents of the file.
     * @function {String} ?
     * @return The contents of injected-code.js
     */
    _loadBaseInjectedCode: function() {
        var io = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
        try  {
            var channel = io.newChannel('chrome://gtdinbox/content/js/gmail/injected-code.js', null, null);
            var input = channel.open();
            var binaryInputStream = Components.classes["@mozilla.org/binaryinputstream;1"].getService(Components.interfaces.nsIBinaryInputStream);
            binaryInputStream.setInputStream(input);

            var parts = [];
            for (var count = input.available(); count; count = input.available()) {
                parts.push(binaryInputStream.readBytes(count));
            }

            binaryInputStream.close();

            return parts.join('');
        } catch (e) {
            glLogger.error('Error loading injected-code.js file:'+"\n"+e, 'glPageWrapper');
        }
    },
    //}
}, {
    // Static Properties {
    /**
     * Dispatched when the page is loaded and ready to interact with
     * @event ?
     */
    UI_READY: 'ui_ready',

    /**
     * Dispatched when gmonkey notifies us of a view change
     * @event ?
     */
    VIEW_CHANGED: 'view_changed',
    //}
});
glPageWrapper.implement(glEventDispatcher); // Add event dispatching support