/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Provides an abstraction layer to GMail searching, with the ability to
 * build up complex search conditions and run the search in a variety of ways.
 *
 * @class ?
 */
var glSearch = glBase.extend({
    // Properties {
    /**
     * The glSearchConditions object for the query
     * @property {glSearchConditions} ?
     */
    conditions: null,

    /**
     * glGmail object
     * @property {glGmail} ?
     */
    _glGmail: null,

    /**
     * The glRequest object
     * @property {glRequest} ?
     */
    _request: null,
    //}

    /**
     * The constructor.
     * Creates a search object and initializes the conditions property.
     * @constructor ?
     * @param {glGmail} gmail - The glGmail object for the page
     */
    constructor: function(gmail) {
        this._glGmail = gmail;
        this.conditions = new glSearchConditions();
    },

    // Methods {
    /**
     * Adds the given search string to the query. Alias for {@link addCondition}.
     * @function {glSearch} ?
     * @paramset Input 1
     * @param {glSearchConditions} condition - The condition to add the query
     * @paramset Input 2
     * @param {String} condition - The condition to add the query
     */
    find: function(condition) {
        return this.addCondition(condition);
    },

    /**
     * Add the given condition to the query
     * @function {glSearch} ?
     * @paramset Input 1
     * @param {glSearchConditions} condition - The condition to add the query
     * @paramset Input 2
     * @param {String} condition - The condition to add the query
     * @return The object, so that it can be chained
     */
    addCondition: function(condition) {
        this.conditions.addCondition(condition);
        return this;
    },

    /**
     * Runs the query, putting it in the searchbox on the page and submitting it
     * @function ?
     */
    run: function() {
        var searchBox = this._glGmail.getUIObject('searchBox');
        searchBox.setSearch(this.getQuery());
        searchBox.submit();
    },

    /**
     * Runs the query in the background
     * @function ?
     * @param {int} start - The start offset (defaults to 0)
     * @param {int} num - The number of rows to return (defaults to 25)
     * @param {optional Object} contextData - Data to be returned with the callback.
     */
    runBackground: function(start, num, contextData) {
        if(typeof start != 'number') start = 0;
        if(typeof num != 'number') num = 25;
        this._reset();

        this._request = new glRequest(this._glGmail, { start: start, num: num, search: 'query', view: 'tl', q: this.getQuery() }, null, contextData);
        this._request.addEventListener(glRequest.COMPLETED, this._backgroundSearchCallback, this);
        this._request.run();
    },

    /**
     * Background search callback
     * @function ?
     */
    _backgroundSearchCallback: function(event) {
        glLogger.log('Search callback called', 'glSearch');

        this.dispatchEvent({name: glSearch.COMPLETED, contextData:event.contextData});
    },
    
    /**
     * Runs the query in the background, but sets the number of results returned
     * to 0 to speed up the request and reduce load on Gmail servers (hopefully).
     * @function ?
     */
    runCount: function() {
        this.runBackground(0, 0);
    },

    /**
     * Returns the number of results returned by the search. This is not the same
     * as the number of actual threads returned that we have data for, which can
     * be calculated by calling <code>getResults().length</code>.
     * @function {int} ?
     * @return The number of results in the search
     */
    getNumberResults: function() {
        if(!this._request) glLogger.error('this._request not defined', 'glSearch');

        var queryData = this._request.data.getQueryData();
        return queryData.results;
    },

    /**
     * Returns an array of the threads returned
     * @function {glThread[]} ?
     * @return The thread data from the search results
     */
    getResults: function() {
        if(!this._request) glLogger.error('this._request not defined', 'glSearch');

        var threads = this._request.data.getThreads();
        return threads;
    },

    /**
     * Returns the query string
     * @function {String} ?
     * @return The query string
     */
    getQuery: function() {
        return this.conditions.toString();
    },

    /**
     * Returns whether or not the given thread matches the search
     * @function {Boolean} ?
     * @param {glThread} thread - The thread to check
     * @return Whether or not it's a match
     */
    matchThread: function(thread) {
        return this.conditions.matchThread(thread);
    },


    /**
     * Cleans up the remnants of the previous search run, preserving the conditions
     * that were built up.
     * @function ?
     */
    _reset: function() {
        glLogger.log('Reseting the glSearch object', 'glSearch');
        if(this._request) {
            this._request.removeEventListener(glRequest.COMPLETED, this._backgroundSearchCallback);
            this._request = null;
        }
    }
    //}
},{
    // Static Properties {
    /**
     * Dispatched when the search is completed
     * @event ?
     */
    COMPLETED: 'completed',
    //}

    // Static Methods {
    /**
     * Turns the given label into a query for that label
     * @function {static String} ?
     * @param {String} label - The label to search for
     * @return Search query string
     */
    l: function(label) {
        return 'label:'+label.toLowerCase().replace(/[\/\s\(\)]/g, '-');
    },
    
    	/**
    	* Get query glSearchConditions that focuses a search on a specific user (who may have multiple email addresses)
    	* ((from:andy@gmail.com OR from:andy@gtdinbox.com) AND (to:andy@gmail.com OR to:andy@gtdinbox.com)) - personal emails
    	* ((from:andy@gmail.com OR from:andy@gtdinbox.com) OR (to:andy@gmail.com OR to:andy@gtdinbox.com)) - email concerning a user
    	* @function {static glSearchConditions} ?
    	* @paramset Single email
    	* @param {String} userids - The username or email address
    	* @param {String} fromOrTo - The join type (AND or OR)
    	* @paramset Single email
    	* @param {String[]} userids - The usernames or email addresses
    	* @param {String} fromOrTo - The join type (AND or OR)
    	* @return A self contained (bracketed) query string
    	*/
	restrictToUser: function(userids, fromOrTo){	// create search block for a given user (set of names or emails)
		if( typeof userids=="string" ) userids = [userids];
		var cFrom = new glSearchOrConditions();
		var cTo = new glSearchOrConditions();
		if( userids instanceof Array ){
			for( var i = 0; i < userids.length; i++ ){
				cFrom.addCondition("from:" + userids[i]);
				cTo.addCondition("to:" + userids[i]);
			}
		}else{
			for( e in userids ){
				cFrom.addCondition("from:" + e);
				cTo.addCondition("to:" + e);
			}
		}

		var c = null;
		if( fromOrTo ){
			c = new glSearchOrConditions();
		}else{
			c = new glSearchAndConditions();
		}
		c.addCondition(cFrom);
		c.addCondition(cTo);
		
		//var joinStr = fromOrTo? "OR" : "AND";
		//return "(" + "(" + fromStr.join(" OR ") + ") " + joinStr + " (" + toStr.join(" OR ") + ")" + ")";
		return c;
	},
	
	/**
	* Get a Gmail-suitable search string for a date
	* @function {static String} ?
	* @param {int} ms - The valueOf of a given date
	* @return A query string for a date
	*/
	searchDate: function(ms){
		var d = new Date(ms);
		return d.getFullYear() + "/" + (d.getMonth()+1) + "/" + d.getDate();
	},
	
	/**
	* Combine labels into a glSearchConditions
	* @function {static glSearchConditions} ?
	* @param {String[]} labels - The labels
	* @param {String} joinType - AND or OR. Defaults to OR.
	* @param {boolean} incBrackets - If true, encapsulate in brackets
	* @return The query string
	*/
	joinLabels: function(labels, joinType, incBrackets){
		// Convert to Gmail label search items:
		var c = null;
		if( !joinType || joinType.toLowerCase()=="or" ){
			c = new glSearchOrConditions();
		}else{
			c = new glSearchAndConditions();
		}

		if(labels) for( var i = 0; i < labels.length; i++ ) c.addCondition(glSearch.l(labels[i]));
		
		return c;
		
		//if( !labels ) return "";
		//for( var i = 0; i < labels.length; i++ ) labels[i] = glSearch.l(labels[i]);
		//var queryStr = labels.join(" " + joinType + " ");
		//return incBrackets? ("(" + queryStr + ")") : queryStr;
	},
    //}
});
glSearch.implement(glEventDispatcher); // Add event dispatching support


/**
 * Container for search conditions
 *
 * @class ?
 */
var glSearchConditions = glBase.extend({
    // Properties {
	/**
	 * Array of the search conditions
     * @property {Array} ?
	 */
	conditions: null,

	/**
	 * Cache of the query string
     * @property {String} ?
	 */
	_queryString: null,
	
    //}

	/**
     * The constructor.
     * @constructor ?
     * @paramset Input 1
     * @param {glSearchConditions} condition - Optional conditions to initialize with
     * @paramset Input 2
     * @param {String} condition - Optional conditions to initialize with
	 */
	constructor: function() {
        if(arguments.length) {
            this.conditions = arguments;
        } else {
            this.conditions = [];
        }
	},

    // Methods {
    /**
     * Adds the given condition
     * @function ?
     * @paramset Input 1
     * @param {glSearchConditions} condition - A condition to add to the current condition
     * @paramset Input 2
     * @param {String} condition - A condition to add to the current condition
     */
    addCondition: function(condition) {
		this.conditions.push(condition);
		this._queryString = null; // Invalidate cache
	},

    /**
     * Converts the condition to a query string
     * @function {String} ?
     * @return The query string
     */
    toString: function() {
		if(this._queryString === null) {
			this._queryString = this._toQueryString();
		}
		return this._queryString;
	},

    /**
     * Returns whether or not the given thread matches the search
     * @function {Boolean} ?
     * @param {glThread} thread - The thread to check
     * @return Whether or not it's a match
     * @throws Impossible match error, if the search condition isn't supported
     */
    matchThread: function(thread, type) {
        // Type is a hidden parameter that determines the type of condition it is - 'and' vs 'or'
        if(typeof type == 'undefined') type = 'and';


        var pass = null;
        for(var i = 0; i < this.conditions.length; i++) {
            // Based on condition type, pass or fail, resetting pass variable afterwards
            if(pass !== null) {
                if(type == 'and' && pass === false) return false;
                if(type == 'or' && pass === true) return true;
            }
            pass = null;

            if(this.conditions[i] instanceof glSearchConditions) {
                pass = this.conditions[i].matchThread(thread);
            } else if(typeof this.conditions[i] == 'string') { // No need to test 'to:' or 'has:attachment', etc. -> only revolves around things that can change. Those only change as new messages arrive (and therefore could be handled; but 'to:' is hard unless messages are available -> which they would be in that use-case)
                if(/^label:/.test(this.conditions[i])) {
					// Must be able to handle the fact label search is not the same as a label (e.g P/GTDInbox -> P-GTDInbox); and there are close cases, e.g. P/GTDInbox and P/GTDInbox/Bugs, that can disrupt simple string tests
					var re = RegExp('(^|[^A-Z0-9-])' + this.conditions[i].substr(6) + '([^A-Z0-9-]|$)', 'i');
					pass = re.test(thread.labels.toString().replace(/[\/\s\(\)]/g, '-')); // TODO Faster to create RE object or to do simple tests iterating over the labels array
                }else if(/^-label:/.test(this.conditions[i])) {
					var re = RegExp('(^|[^A-Z0-9-])' + this.conditions[i].substr(7) + '([^A-Z0-9-]|$)', 'i');
					pass = !re.test(thread.labels.toString().replace(/[\/\s\(\)]/g, '-'));
				}else if(/^has:attachment/.test(this.conditions[i])){
					pass = !!thread.attachmentsCSV;
		                /*
		                // Cannot be done as without full message loading (too costly) we cannot differentiate from and to
				}else if(/^to:/.test(this.conditions[i])){
					pass = thread.hasContact( this.conditions[i].substr(3) );
		                	glLogger.log('test include to:');
					// TODO Grab getThreadDetails, but without requesting an ajax download; and if messages present, test for to:Andy (would include Andy Mitchell), to:Mitchell (same), and to:andy@gtdinbox.com
				}else if(/^from:/.test(this.conditions[i])){
		                	glLogger.log('test include from:');
					// TODO Grab getThreadDetails, but without requesting an ajax download; and if messages present, test for to:Andy (would include Andy Mitchell), to:Mitchell (same), and to:andy@gtdinbox.com
				*/
				} else {
                    throw new Error('Impossible match');
                }
            } else {
                throw new Error('Impossible match');
            }
        }

        // Made it to the end, so the final test determines matching
        return pass;
    },

    /**
     * Converts the condition to a query string using the given seperator
     * @function {String} ?
     * @param {String} seperator - The optional seperator to place between conditions
     * @return The query string
     */
    _toQueryString: function(seperator) {
        var queryString = seperator ? "(" : "";

        for(var i = 0; i < this.conditions.length; i++) {
            if(!this.conditions[i]) continue;

            if(typeof this.conditions[i] == 'string') {
                queryString += this.conditions[i];
            } else if(this.conditions[i].toString) {
                queryString += this.conditions[i].toString();
            } else {
                glLogger.error('Invalid condition', 'glSearchConditions');
            }

            if(i < this.conditions.length-1) queryString += seperator ? seperator : " ";
        }

        if(seperator) queryString += ")";

        return queryString;
	},
    //}
});

/**
 * Container for search conditions which joins them with an "AND"
 *
 * @class ? @extends glSearchConditions
 */
var glSearchAndConditions = glSearchConditions.extend({
    // Methods {
    /**
     * @function {Boolean} ?
     * @inheritdesc
     */
    matchThread: function(thread) {
        return this.base(thread, 'and');
    },

    /**
     * Converts the condition to a query string
     * @function {String} ?
     * @return The query string
     */
	toString: function() {
		if(this._queryString === null) {
			this._queryString = this._toQueryString(' AND ');
		}
		return this._queryString;
	}
    //}
});

/**
 * Container for search conditions which joins them with an "OR"
 *
 * @class ? @extends glSearchConditions
 */
var glSearchOrConditions = glSearchConditions.extend({
    // Methods {
    /**
     * @function {Boolean} ?
     * @inheritdesc
     */
    matchThread: function(thread) {
        return this.base(thread, 'or');
    },

    /**
     * Converts the condition to a query string
     * @function {String} ?
     * @return The query string
     */
	toString: function() {
		if(this._queryString === null) {
			this._queryString = this._toQueryString(' OR ');
		}
		return this._queryString;
	}
    //}
});
