/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Central cache for storing lists of threads.
 * Automatically populated when the view changes in Gmail (e.g. Inbox, Sent Items, Search Results).
 * TODO - make it possible to get threads even if there's a cache miss - download them, and issue 'update' events so a listener can check to see if their threads have arrived yet
 *
 * @class ? @extends glBase
 */

var glThreadListDataCache = glBase.extend({

	_gmail: null,
	
	/**
     * Hash containing cached thread list data. The key is the URL that the request
     * was made to and the value is a hash containing the data (key "data"), and
     * a boolean flag (key "doDelete") that will trigger a delete of the data
     * the next time the view changes.
     * @property {Object} ?
     */
    _threadListDataCache: null,
    
	/**
	* The Constructor
	* @function ?
	* @param {glGmail} gmail - the glgmail instance for the page
	* @param {glDataObject} viewData - the glDataObject for Gmail's VIEW_DATA
	*/
	constructor: function(gmail, viewData){
		this._gmail = gmail;
		this._threadListDataCache = {};
		
		if( this._gmail.environment.trafficWatcher ){
			this._gmail.environment.trafficWatcher.addEventListener(glTrafficWatcher.THREAD_LIST_REQUEST_COMPLETED, this._handleThreadListRequest, this);
		}
		
		if( viewData ){
			// Create fake url and populate _threadListDataCache for the inbox from the view data
	        var threads = viewData.getThreads();
	        var fakeURL = 'mail?search=inbox&start=0&num='+threads.length;
	        this._threadListDataCache[fakeURL] = {data:viewData};
		}
		
		
        this._gmail.page.addEventListener(glPageWrapper.UI_READY, this._baseUIReady, this);
		
	},
	
	/**
     * The UI_READY callback for the glGmail class. Currently just populates
     * the canvasDocument property.
     * @function ?
     * @param {Event} evt - The glPageWrapper.UI_READY event
     */
	_baseUIReady: function(){
		glLogger.log('LOAD BASEUIREADY');
    	this._gmail.actions.addEventListener(glActionManager.THREAD_CHANGED, this._threadChanged, this);	
	},
	
	/**
     * Checks the internal thread list data cache to see if it has data matching
     * the request and returns it.
     * @function {Array} ?
     * @param {String} viewHash - The hash portion of the URL for the current view ('#search/Apple+computers' for example)
     * @param {int} startOffset - The start offset of the thread data
     * @param {int} endOffset - The end offset of the thread data
     * @return The threads that it found - empty array if none found
     */
    getThreadListData: function(viewHash, startOffset, endOffset) {
        glLogger.log('Getting thread list data...', 'glEnvironment');


		// Strip pagination from URL (#inbox/p2 => #inbox)
		if( /\/p\d+$/.test(viewHash) ) viewHash = viewHash.replace(/\/p\d+$/, '');
		

		
        // Build regex for matching the URL in the cache from the viewHash argument
		// Also escape search string. e.g. has:attachment => has%3Aattachment
        var urlRegex = null;
        if(viewHash.substr(0, 8) == '#search/') {
            glLogger.log('Regex: "'+'[?&]q='+viewHash.substr(8)+'"', 'glEnvironment');
            urlRegex = new RegExp('[?&]q='+glBase.prepRE(escape(viewHash.substr(8))));
        } else {
            urlRegex = new RegExp('[?&]search='+glBase.prepRE(escape(viewHash.substr(1))));
        }

        var threads = [];
        for(var url in this._threadListDataCache) {
            if(urlRegex.test(url)) {
                glLogger.log('Found thread list data "'+url+'" - checking offsets: '+startOffset+' - '+endOffset, 'glEnvironment');

                // Get start and end offset out of url
                var urlStartOffset = parseInt(/[?&]start=(\d+)/.exec(url)[1]);
                var urlEndOffset = urlStartOffset + parseInt(/[?&]num=(\d+)/.exec(url)[1]) - 1;

                // Check if in range
                if(urlStartOffset > startOffset || urlEndOffset < endOffset) continue;

                // In range so build thread list and return
                glLogger.log('Offsets check out so building thread array', 'glEnvironment');
                var data = this._threadListDataCache[url].data.getThreads();
                
                //this._gmail.doc.defaultView.wrappedJSObject.top.console.log(this._threadListDataCache[url].data);

                var startIndex = startOffset - urlStartOffset;
                var endIndex = endOffset - urlStartOffset;
                if(endIndex > data.length) endIndex = data.length;

                for(var i = startIndex; i <= endIndex; i++) {
                    threads.push(data[i]);
                }
                break;
            } else {
                glLogger.log('Url failed: "'+url+'"', 'glEnvironment');
            }
        }

        return threads;
    },
	
	/*
	* Search thread caches for first thread with this ID, and return it
	* @function {glThread} ?
	* @param {String} threadId
	* @return The thread object with threadId
	*/
	getThreadById: function(threadId){
		for( key in this._threadListDataCache ){
			var threads = this._threadListDataCache[key].data.getThreads();
			for( var i = 0; i < threads.length; i++ ){
				if( threads[i].hasId(threadId) ){
					return threads[i];
				}
			}
		}
	},
	
	/**
     * Event handler for {@link glTrafficWatcher.THREAD_LIST_REQUEST_RECEIVED}
     * events. Adds the response data to the thread list cache.
     * @function ?
     * @param {Event} evt - The THREAD_LIST_REQUEST_RECEIVED event
     */
    _handleThreadListRequest: function(evt) {
	
		for( key in this._threadListDataCache ) this._threadListDataCache[key].mostRecent = false;
		if( this._threadListDataCache[evt.url] ) delete this._threadListDataCache[evt.url];
        this._threadListDataCache[evt.url] = {data: evt.responseData, doDelete: false, mostRecent: true, created: (new Date()).valueOf()};
    },

   /**
     * Callback for the {@link glActionManager.THREAD_CHANGED} event. Updates all
     * stored threads from the event and removes any that no longer fit the search
     * criteria.
     * @function ?
     * @param {Event} evt - The thread changed event
     */
	_threadChanged: function(evt) {
		// Find instances of the thread
		for( key in this._threadListDataCache ){
			var threads = this._threadListDataCache[key].data.getThreads();
			for( var i = threads.length - 1; i>=0; i-- ){
				if( threads[i].threadId==evt.thread ){
					
					threads[i].updateFromEvent(evt);
										
					if( evt.action==glActionManager.ARCHIVE ){				
						if( key.indexOf('inbox')>-1 ){	// TODO Can we improve precision of this. Do we need to?
							threads.splice(i, 1);
						}
						
					}else if( threads[i].isDeleted || threads[i].hasLabel("^s") ){	// TODO ^s is spam. Could be better introduced into glThread; partly by making glActionManager aware of it (and able to watch it)
						threads.splice(i,1);
					}
				}
			}
		}
	},



});