/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * The thread list view abstraction layer.
 *
 * @class ? @extends glUIView
 */
var glThreadListView = glUIView.extend({
    // Properties {
    /**
     * The {@link glActionsHeader} object for the page
     * @property {glActionsHeader} ?
     */
    actionsHeader: null,



  
    //}

    // Methods {
    /**
     * Connects the element to the given glGmail object's page. Called automatically
     * by {@link glGmail.getUIObject}.
     * @function ?
     * @param {glGmail} gmail - The glGmail object for the page
     */
    connect: function(gmail) {
        glLogger.log('Connecting glThreadListView object to view', 'glThreadListView');

        this.base(gmail);

        this.actionsHeader = gmail.getUIObject('actionsHeader');
    },

    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    disconnect: function() {
        glLogger.log('Disconnecting glThreadListView from view', 'glThreadListView');
        this.actionsHeader.disconnect();
        this.base();
    },

    /**
     * Returns the starting offset, ending offset, and number of results in a hash
     * @function {Object} ?
     * @return Pagination information
     * @... {int} startingOffset - The starting thread offset
     * @... {int} endingOffset - The ending thread offset
     * @... {int} results - The number of threads in the result
     */
    getPaginationInfo: function() {
        // Get span that contains the info
        var result = this._glGmail.canvasDocument.evaluate(".//span/b", this.actionsHeader.pagination, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null);
        var span = result.singleNodeValue.parentNode;
        var bNodes = span.getElementsByTagName('b');

        return {
            startingOffset: parseInt(bNodes[0].textContent),
            endingOffset: parseInt(bNodes[1].textContent),
            results: parseInt(bNodes[2].textContent),
        };
    },

    /**
     * Returns the data for the thread at the given offset
     * @function {glThread} ?
     * @param {int} offset - The offset in the current view (ignoring pagination)
     * @return The thread data - null if not available
     */
    getThread: function(offset) {
        var threads = this.getThreads();
        if(threads.length > offset) {
            return threads[offset];
        } else {
            return null;
        }
    },

    /**
     * Return the thread data corresponding to the current view if it can get it
     * @function {glThread[]} ?
     * @return The thread data
     */
    getThreads: function() {
            var paginationInfo = this.getPaginationInfo();
            return this._glGmail.environment.threadListDataCache.getThreadListData(this._glGmail.doc.location.hash, paginationInfo.startingOffset-1, paginationInfo.endingOffset-1);
    },
    
    	/**
    	* Return array of TABLE elements. The first is always the main thread list; remaining are 'multiple inbox' tables.
    	* @function {Element[]} ?
    	* @return Array of TABLE elements that represent thread lists
    	*/
    	getThreadListTables: function(){
    		var doc = this._glGmail.canvasDocument;
			var result = doc.evaluate( "//div[contains(@class,'glViewRoot')]//table[contains(@class, 'RnTrV')]", doc.body, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
		
			var tables = [];
			for( var i = 0; i < result.snapshotLength; i++ ) tables.push( result.snapshotItem(i) );
			return tables;
    	},
    
    	/**
    	* Return xpath iterator of TD elements - the last in each row TR in the thread lists
    	* @function ?
    	* @return XPath iterator of TD elements
    	*/
	getLastTDInRows: function(){
		var doc = this._glGmail.canvasDocument;
		var result = doc.evaluate( "//div[contains(@class,'glViewRoot')]//tr[starts-with(@class,'zA')]//td[last()]", doc.body, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
		if( result.snapshotLength==0 ) result = doc.evaluate( "//div[contains(@class,'glViewRoot')]//div[contains(@class,'EGSDee')]/div/div[contains(@class,'AH8jVb')]//table[contains(@class,'N2mZ7b')]//tr[starts-with(@class,'xweT7d')]//td[last()]", doc.body, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
		return result;
	},
	
	/** 
	* Return offset of row element provided; starting at 0 for the first row in view. Ignores pagination.
	* function {int} ?
	* @param {Element} elRow - The row element (or any element within the TR) to get the offset for
	* @return The offset for provided row
	*/
	getOffset: function(elRow){
		// Check it's a TR
		while( elRow.parentNode && elRow.nodeName!="TR" ) elRow = elRow.parentNode;
		if( elRow.nodeName!="TR" ) return null;
		
		// Now count back it's position
		var offset = 0;
		while( elRow.previousSibling ){
			offset++;
			elRow = elRow.previousSibling;
		}
		return offset;
	},
	
	/**
	* Uncheck all visible rows
	* function ?
	*/
	uncheckRows: function(){
		this.checkedRows([]);
	},
	
	/**
	* Either retrieve index of selected rows; or set selected rows based on index
	* @function {Boolean[]} ?
	* @param {Boolean[]} checked - If present, set checked rows according to array e.g. [True, False, False, True...]
	* @return If no 'checked', return array of checked rows
	*/
	checkedRows: function(checked){
		var elTL = this._glGmail.page.getViewRoot();
		var elInputs = elTL.getElementsByTagName("INPUT");
		
		if( typeof checked=='undefined' ){
			var checked = [];
			var j = 0;
			for( var i = 0; i < elInputs.length; i++ ){
				if( elInputs[i].getAttribute("type").toLowerCase()=="checkbox" ){
					if( elInputs[i].checked ) checked[j++] = true
				}
			}
			return checked;
		}else{	
			var j = 0;
			for( var i = 0; i < elInputs.length; i++ ){
				if( elInputs[i].getAttribute("type").toLowerCase()=="checkbox" ){
					j++;
					if( (checked[j] && !elInputs[i].checked) || (!checked[j] && elInputs[i].checked) ) glUIWidget.simulateMouseEvent(elInputs[i], "click"); // Toggle
				}
			}
		}
	},
	
    //}
});
