/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Provides an abstraction layer for accessing stored data related to a specific
 * gmail account. This includes preferences and saved data.
 *
 * @class ?
 */
var gtdAccountData = glBase.extend({
    // Properties {

    /**
     * Reference to the gtdMainPagePlugin for the page
     * @property {gtdMainPagePlugin}
     */
    _gtd: null,

    /**
     * The e-mail address of the account
     * @property {String} ?
     */
    _account: null,

    /**
     * The account preferences
     * @property {Object} ?
     */
    _prefs: null,

    /**
     * Whether the preferences have been changed since the last save
     * @property {Boolean} ?
     */
    _prefsChanged: false,

    /**
     * Periodic save timeout identifier
     * @property {int} ?
     */
    _periodicSaveTimeout: false,
    
    /**
     * True if last attempt to connect GTDInbox server was successful.
     * @property {Boolean} ?
     */
    _synced: false,
    
    /**
     * True if account has been properly loaded
     * @property {Boolean} ?
     */
    _loaded: false,
    
    /**
     * True if currently writing to the server. Try to avoid multiple simultaneous calls. 
     * @property {Boolean} ?
     */
    _serverWriteLock: false,
    
    /**
     * Reference to all requests made so they can be cleaned up
     * @property {gtdAccountRequest[]} ?
     */
    _requests: null,
    
    //}

    // Methods {
    // General {
    /**
     * Shuts down the gtdAccountData object and saves everything to disk
     * @function ?
     */
    shutdown: function() {
	for( var i = 0; i < this._requests.length; i++ ){
	    this._requests[i].reset();
	}
        this.savePrefs();
        clearTimeout(this._periodicSaveTimeout);
    },
    
    /**
     * Find gtdAccountRequest for this evt.target and clean it up
     * @function {} ?
     */
    _clearOutRequest: function(evt){
	for( var i = this._requests.length -1; i>=0; i-- ){
	    if( this._requests[i]===evt.target ){
		this._requests[i].reset();
		this._requests.splice(i, 1);
	    }
	}
    },

    //}
    
    
    // Load Specific {

    /**
     * Called by {@link retrieveForAccount}, it sets the _account variable
     * @function ?
     * @param {String} account - The account e-mail address to load up settings for
     */
    _loadForAccount: function(account) {
        this._account = account;
    },
    
    /**
     * Start the load process
     * @function ?
     */
    load: function(){
	// Try to retrieve latest settings
	var req = gtdAccountRequest.createReadSettingsRequest(this._gtd, this._account);
	this._requests.push(req);
	req.addEventListener(gtdAccountRequest.SUCCESS, this._loadForAccountCallback, this);
	req.addEventListener(gtdAccountRequest.FAIL, this._loadForAccountCallback, this);
	req.addEventListener(gtdAccountRequest.WRONG_PASSWORD, this._loadForAccountCallback, this);
	req.addEventListener(gtdAccountRequest.NO_ACCOUNT, this._loadForAccountCallback, this);
	req.run();
    },
    
    /**
     * Notification the ReadSettings request has run
     * @function {} ?
     */
    _loadForAccountCallback: function(evt){

	switch(evt.name){
	    case gtdAccountRequest.SUCCESS:
		// Ensure cache any unexported changes (prior to potentially overwriting with a server version):
		var uploadQueue = this._gtd.getPref("uploadQueue");    // Format: [{key:String, value:String, timestamp:int}]
		// Use retrieved settings as local prefs:
		if( evt.result.settings ){
		    var prefManager = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
		    var prefString = gtdAccountData.JSON.encode(evt.result.settings);
		    prefManager.setCharPref("extensions.gtdinbox."+this._account, prefString);
		}
		// Read in these new settings
		this._loadForAccountUseLocal();
		// Apply uploadQueue to new local prefs
		if( uploadQueue ){
		    for( var i = 0; i < uploadQueue.length; i++ ){
			this.setPref(uploadQueue[i].key, uploadQueue[i].value, true);	// Nb: does not re-add these items to uploadQueue in setPref
		    }
		    if( uploadQueue.length>0 ){
			// There were outstanding changes. Upload them.
			this.savePrefs();
		    }
		}
		this._synced = true;
		this.dispatchEvent({name: gtdAccountData.LOADED});
		break;
	    case gtdAccountRequest.FAIL:
		// Could not read from server. Use local version
		this._synced = false;
		this._loadForAccountUseLocal();
		this.dispatchEvent({name: gtdAccountData.LOADED});
		break;
	    case gtdAccountRequest.WRONG_PASSWORD:
		// Need to discern whether they have an account or not
		if( (typeof this._lastFailedPassword=="undefined") || this._lastFailedPassword!=evt.target.password ){	
		    // Show Change Password
		    this._lastFailedPassword = evt.target.password;
		    this._changePassword = this._gtd.getUIObject('changePassword');
		    this._changePassword.addEventListener(gtdChangePassword.TRY_AGAIN, this._loadChangePasswordCallback, this);
		    this._changePassword.addEventListener(gtdChangePassword.CANCEL, _loadChangePasswordCallback, this);
		    this._changePassword.show();
		}else{
		    // Silently fail, use local
		    this._synced = false;
		    this._loadForAccountUseLocal();
		    this.dispatchEvent({name: gtdAccountData.LOADED});
		}
		/*
		var req = gtdAccountRequest.createHasAccountRequest(this._gtd, this._account);
		this._requests.push(req);
		req.addEventListener(gtdAccountRequest.SUCCESS, this._wrongPasswordHasAccountCallback, this);
		req.addEventListener(gtdAccountRequest.FAIL, this._wrongPasswordHasAccountCallback, this);
		req.run();
		*/
		break;
	    case gtdAccountRequest.NO_ACCOUNT:
		// Need to show install screen
		this._installer = this._gtd.getUIObject('installer');
		this._installer.addEventListener(gtdInstaller.INSTALLED, this._loadInstallerCallback, this);
		this._installer.addEventListener(gtdInstaller.CANCEL, _loadInstallerCallback, this);
		this._installer.show();
		break;
	}
	this._clearOutRequest(evt);
    },
    
    /**
     * Read in settings from Firefox preferences (which could have been just updated from server).
     * Migrate any updates in the code (which will be uploaded back to the server).
     * @function {}?
     */
    _loadForAccountUseLocal: function(){
	this._loadPrefs();
        this._migratePrefs();
	this._loaded = true;
    },
    
    /**
     * After a wrong password, check to see if the user even has an account.
     * If they do, ask them to confirm their password.
     * If they do not, show install box
     * @function {} ?
     */
    /*
    _wrongPasswordHasAccountCallback: function(aEvt){
	switch(aEvt.name){
	    case gtdAccountRequest.SUCCESS:
		if( aEvt.result.hasAccount ){
		    // They had a wrong password. If first time, show Change Password
		    if( (typeof this._lastFailedPassword=="undefined") || this._lastFailedPassword!=aEvt.target.password ){	
			// Show Change Password
			this._lastFailedPassword = aEvt.target.password;
			this._changePassword = this._gtd.getUIObject('changePassword');
			this._changePassword.addEventListener(gtdChangePassword.TRY_AGAIN, this._loadChangePasswordCallback, this);
			this._changePassword.addEventListener(gtdChangePassword.CANCEL, _loadChangePasswordCallback, this);
			this._changePassword.show();
		    }else{
			// Silently fail, use local
			this._synced = false;
			this._loadForAccountUseLocal();
	    		this.dispatchEvent({name: gtdAccountData.LOADED});
		    }
		}else{
		    // Need to show install screen
		    this._installer = this._gtd.getUIObject('installer');
		    this._installer.addEventListener(gtdInstaller.INSTALLED, this._loadInstallerCallback, this);
		    this._installer.addEventListener(gtdInstaller.CANCEL, _loadInstallerCallback, this);
		    this._installer.show();
		}
		break;
	    case gtdAccountRequest.FAIL:
		// Unexpected server fail, possible but unlikely (as to get this far must have succeeded in reading settings)	    
		this._synced = false;
		this._loadForAccountUseLocal();
		this.dispatchEvent({name: gtdAccountData.LOADED});
		break;
	}
	this._clearOutRequest(aEvt);
    },
    */
    
    _loadChangePasswordCallback: function(evt){
	this._changePassword.removeEventListener(gtdChangePassword.TRY_AGAIN, this._loadChangePasswordCallback, this);
	this._changePassword.removeEventListener(gtdChangePassword.CANCEL, this._loadChangePasswordCallback, this);
	switch( evt.name ){
	    case gtdChangePassword.TRY_AGAIN:
		// Supplied a new password, try loading preferences again
		this._loadForAccount(this._account);
		break;
	    case gtdChangePassword.CANCEL:
		// Use local version
		this._synced = false;
		this._loadForAccountUseLocal();
		this.dispatchEvent({name: gtdAccountData.LOADED});
		break;
	}
    },
    
    
    _loadInstallerCallback: function(evt){
	// TODO What happens if we install an account, then server fails and doesn't recognise password or settings are not committed? 
	this._installer.removeEventListener(gtdInstaller.INSTALLED, this._loadInstallerCallback, this);
	this._installer.removeEventListener(gtdInstaller.CANCEL, this._loadInstallerCallback, this);
	switch( evt.name ){
	    case gtdInstaller.INSTALLED:
		// Supplied a new password, try loading preferences again
		this._loadForAccount(this._account);
		break;
	    case gtdInstaller.CANCEL:
		// Use local version
		this._synced = false;
		this._loadForAccountUseLocal();
		this.dispatchEvent({name: gtdAccountData.LOADED});
		break;
	}
    },
    
    
    //}






    // Pref Specific {
    /**
     * Returns the value of the given preference
     * @function {Object} ?
     * @param {String} pref - The preference string
     * @param {boolean} suppressNotFoundError - Optional. Do not throw error if not found.
     * @return The value of the preference
     */
    getPref: function(pref, suppressNotFoundError) {
        var prefPath = pref.split('.');
        var wantedPref = this._prefs;
        while(prefPath.length) {
            var next = prefPath.shift();
            if(typeof wantedPref[next] != 'undefined') {
                wantedPref = wantedPref[next];
            } else {
            	if( suppressNotFoundError ){
            		return null;
            	}else{
	                glLogger.error('No preference: "'+pref+'"', 'gtdAccountData');
	        }
            }
        }

        return wantedPref;
    },

    /**
     * Sets the preference to the given value
     * @function ?
     * @param {String} pref - The preference string
     * @param {Object} value - The value to set the preference to
     * @param {Boolean} doNotUpload - If true, do not write it to the uploadQueue
     */
    setPref: function(pref, value, doNotUpload) {
        var prefPath = pref.split('.');
        var base = this._prefs;
        while(prefPath.length > 1) {
            var currentStep = prefPath.shift();
            if(typeof base[currentStep] == 'undefined') base[currentStep] = {};

            base = base[currentStep];
        }
	if( value===null ){
	    delete base[prefPath[0]];
	}else{
	    base[prefPath[0]] = value;
	}
	
	if( !doNotUpload ){
	    // Write as a transaction
	    if( !this._prefs.uploadQueue ) this._prefs.uploadQueue = [];
	    this._prefs.uploadQueue.push({key:pref, value:value, timestamp:(new Date()).valueOf()});
	}
	
        this._prefsChanged = true;
    },

    /**
     * Serializes the prefs and saves them
     * @function ?
     * @param {Boolean} doNotUpload - If true, do not commit anything to the server
     */
    savePrefs: function(){
        if(this._prefsChanged) {
            var prefManager = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
            var prefString = gtdAccountData.JSON.encode(this._prefs);
            prefManager.setCharPref("extensions.gtdinbox."+this._account, prefString);
        }
        this._prefsChanged = false;

	if( !this._serverWriteLock && this._prefs.uploadQueue && this._prefs.uploadQueue.length>0 ){ // "DO NOT RUN"=="REALLY DON'T RUN" && 
	    // Commit changes to server
	    
	    this._serverWriteLock = true;
	    var uploadQueue = [];
	    for( var i = 0; i < this._prefs.uploadQueue.length; i++ ) uploadQueue.push(this._prefs.uploadQueue[i]);
	    var req = gtdAccountRequest.createWriteSettingsRequest(this._gtd, this._account, uploadQueue);
	    req.timestamp = (new Date()).valueOf();
	    this._requests.push(req);
	    req.addEventListener(gtdAccountRequest.SUCCESS, this._savePrefsCallback, this);
	    req.addEventListener(gtdAccountRequest.FAIL, this._savePrefsCallback, this);
	    req.addEventListener(gtdAccountRequest.WRONG_PASSWORD, this._savePrefsCallback, this);
	    req.addEventListener(gtdAccountRequest.NO_ACCOUNT, this._savePrefsCallback, this);
	    req.run();
	}

        //this._saveAgainAfterTimeout();	// TODO - Stephen, why was this added? Not sure of relevance. Potentially messy in light of server-uploading.
    },
    
    /**
     * Notification of complete following our attempt to write settings to the server
     * @function {} ?
     */
    _savePrefsCallback: function(evt){
	
	
	switch(evt.name){
	    case gtdAccountRequest.SUCCESS:
		// Remove everything in the queue prior to the timestamp of this event
		for( var i = this._prefs.uploadQueue.length - 1; i>=0; i-- ){
		    if( this._prefs.uploadQueue.timestamp < evt.target.timestamp ){
			this._prefs.uploadQueue.splice(i, 1);
		    }
		}
		// Check sync
		if( !this._synced ) this.dispatchEvent({name: gtdAccountData.SYNC_CHANGED});
		this._synced = true;
		// Commit the changes to the local store - and upload again if anything has happened since
		this._serverWriteLock = false;
		this.savePrefs();
		break;
	    case gtdAccountRequest.FAIL:
	    case gtdAccountRequest.WRONG_PASSWORD:
	    case gtdAccountRequest.NO_ACCOUNT:
		// We can fail safely because transactions can be updated later.
		this._synced = false;
		this.dispatchEvent({name: gtdAccountData.SYNC_CHANGED}); 
		break;
	}
	
	this._clearOutRequest(evt);
    },
    
    

    /**
     * Loads and unserializes the account's prefs
     * @function ?
     */
    _loadPrefs: function() {
        if(this._prefs != null) glLogger.error('_loadPrefs can only be called once', 'gtdAccountData');

        glLogger.log('Loading prefs...', 'gtdAccountData');
        var prefManager = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
        try {
            var prefString = prefManager.getCharPref("extensions.gtdinbox."+this._account);
            this._prefs = gtdAccountData.JSON.decode(prefString);
        } catch(e) {
            // Prefs not yet set, so set this._prefs to default
            this._prefs = gtdAccountData.getDefaultPrefs();
            this._prefsChanged = true;
        }
        glLogger.log('Loaded prefs', 'gtdAccountData');

        this._saveAgainAfterTimeout();
    },

    /**
     * Performs migrations on the preferences if they are below the current version
     * @function ?
     */
    _migratePrefs: function() {
        /* Put in case statements for each released version and then it will perform
         * all migrations necessary after that version to get to the current version
         * so long as "break" statements are removed.
         */
        var initialVersion = this._prefs.version;
        switch(this._prefs.version) {
            case '1.0':
		this.setPref('components.traffic_watcher', {enabled:true});
                //this._prefs.components.traffic_watcher = { enabled: true };
		this.setPref('version', '1.0.1');
                //this._prefs.version = '1.0.1';
            case '1.0.1':
                var temp = ['next_action', 'action', 'waiting_on', 'some_day', 'finished'];
		this.setPref('labels.types.statuses.core', {});
		for(var i = 0; i < temp.length; i++) {
		    this.setPref('labels.types.statuses.core.'+temp[i],this._prefs.labels.types.statuses[temp[i]]);
		    this.setPref('labels.types.statuses.'+temp[i],null);
		}
		this.setPref('version', '1.0.2');
		/*
                this._prefs.labels.types.statuses.core = {};
                for(var i = 0; i < temp.length; i++) {
                    this._prefs.labels.types.statuses.core[temp[i]] = this._prefs.labels.types.statuses[temp[i]];
                    delete this._prefs.labels.types.statuses[temp[i]];
                }
                this._prefs.version = '1.0.2';
		*/
            case '1.0.2':
            	if( !this._prefs.components ) this.setPref('components', {});
            	var temp = ['compose_personal','search_button','dashboard'];
            	for( var i = 0; i < temp.length; i++ ){
		    if( !this._prefs[temp[i]] ) this.setPref(temp[i], {enabled:true});
		}
            	if( !this._prefs.label_box ) this.setPref('label_box', {expanded: ['statuses', 'projects', 'contexts', 'misc']});
            	if( !this._prefs.general ) this.setPref('general', {centre_popup:true});
		this.setPref('version', '1.0.3');
	    case '1.0.3':
		this.setPref('components.traffic_watcher', null);
		this.setPref('version', '1.0.4');
            case 'fallthrough':
                // Always called if a migration takes place - do final processing here
                this._prefsChanged = true;
                break;
        }
	
	if( this._prefsChanged ){
	    this.savePrefs();
	}

        if(initialVersion != this._prefs.version) {
            glLogger.log('Migrated prefs from "'+initialVersion+'" to "'+this._prefs.version+'"', 'gtdAccountData');
        }
    },

    /**
     * Calls save again after timeout if greater than zero.
     * Checks if the auto_save_frequency pref is greater than zero and then sets
     * a timeout to save again after the prefered time has passed, if it is
     * greater.
     * @function ?
     */
    _saveAgainAfterTimeout: function() {
        if(!this._prefs) return;
        if(this._prefs.internal.auto_save_frequency > 0) {
            var me = this;
            clearTimeout(this._periodicSaveTimeout);
            this._periodicSaveTimeout = setTimeout(function() { me.savePrefs(); }, this._prefs.internal.auto_save_frequency);
        }
    },
    
    
    	/**
    	* Returns a Status object derived from the preferences - for tidier code.
    	* @function ?
    	* @param {boolean} incReverseLookup - If true, include a .reverseLookup object to find the name of 
    	*/
	getStatusData: function(incReverseLookup){
		var typeSettings = this.getPref('labels.types');
		
		var status = {};	// Cache for tidier code
		status.prefix = typeSettings.statuses.prefix;
		status["today"] = status.prefix + typeSettings.statuses.core.today;
		status["next_action"] = status.prefix + typeSettings.statuses.core.next_action;
		status["action"] = status.prefix + typeSettings.statuses.core.action;
		status["waiting_on"] = status.prefix + typeSettings.statuses.core.waiting_on;
		status["some_day"] = status.prefix + typeSettings.statuses.core.some_day;
		status["finished"] = status.prefix + typeSettings.statuses.core.finished;
	
		if( incReverseLookup ){
			status.reverseLookup = {};
			status.reverseLookup[status["today"]] = "today";
			status.reverseLookup[status["next_action"]] = "next_action";
			status.reverseLookup[status["action"]] = "action";
			status.reverseLookup[status["waiting_on"]] = "waiting_on";
			status.reverseLookup[status["some_day"]] = "some_day";
			status.reverseLookup[status["finished"]] = "finished";
		}
		return status;
	},
	
	
	
    //}
    //}
},{
    // Static Properties {
    /**
     * The default properties object
     * @property {static Object} ?
     */
    DEFAULT_PREFS: {
        version: '1.0.3',
        labels: {
            types: {
                statuses: { prefix: 'S/', core: {today: 'Today', next_action: 'Next Action', action: 'Action', waiting_on: 'Waiting On', some_day: 'Some Day', finished: 'Finished'} },
                projects: { prefix: 'P/' },
                contexts: { prefix: 'C/' },
                references: { prefix: 'R/' },
                old: { prefix: 'Old/' },
                misc: { prefix: '' },
            },
            star_for_next_action: false,
            only_one_status: true,
        },
        reviews: [
            {id:'lessThanOneWeek', categoryKey:'Dashboard.Reviews.Categories.AgingActions', titleKey:'Dashboard.Reviews.LessThanOneWeek', search:'(<?=DefaultStatus.NextAction> OR <?=DefaultStatus.Action>) after:<?=DaysAgo.7> -<?=DefaultStatus.Finished>', builtin:true},
            {id:'moreThanOneWeekLessThanOneMonth', categoryKey:'Dashboard.Reviews.Categories.AgingActions', titleKey:'Dashboard.Reviews.MoreThanOneWeekLessThanOneMonth', search:'(<?=DefaultStatus.NextAction> OR <?=DefaultStatus.Action>) before:<?=DaysAgo.7> after:<?=MonthsAgo.1> -<?=DefaultStatus.Finished>', builtin:true},
            {id:'moreThanOneMonth', categoryKey:'Dashboard.Reviews.Categories.AgingActions', titleKey:'Dashboard.Reviews.MoreThanOneMonth', search:'(<?=DefaultStatus.NextAction> OR <?=DefaultStatus.Action>) before:<?=MonthsAgo.1> -<?=DefaultStatus.Finished>', builtin:true},
            {id:'finishedLastThreeMonths', categoryKey:'Dashboard.Reviews.Categories.RecentComplete', titleKey:'Dashboard.Reviews.FinishedLastThreeMonths', search:'before:<?=MonthsAgo.3> <?=DefaultStatus.Finished>', builtin:true},
        ],
        components: {
            compose_personal: { enabled: true },
            search_button: { enabled: true },
            dashboard: { enabled: true },
            label_box: {
                expanded: ['statuses', 'projects', 'contexts', 'misc']
            },

        },
        general: {
        	centre_popup: true,
        },
        internal: {
            auto_save_frequency: 5*3600*1000,	// TODO disable autosave by default
        },
    },

    /**
     * A reference to the nsIJSON interface
     * @property {static nsIJSON} ?
     */
    JSON: Components.classes["@mozilla.org/dom/json;1"].createInstance(Components.interfaces.nsIJSON),
    //}

    // Static Methods {
    /**
     * Clones the default prefs and returns the clone
     * @function {static Object} ?
     * @return The default prefs as an object
     */
    getDefaultPrefs: function() {
        var json = gtdAccountData.JSON;
        return json.decode(json.encode(gtdAccountData.DEFAULT_PREFS));
    },

    /**
     * Returns an account data object for the given account
     * @function {static gtdAccountData} ?
     * @param {gtdMainPagePlugin} gtd - 
     * @param {String} account - The account e-mail address to load up settings for
     * @return An initialized gtdAccountData object
     */
    retrieveForAccount: function(gtd, account) {
        glLogger.log('Loading account data for "'+account+'"', 'gtdAccountData');
        var accountData = new gtdAccountData();
	accountData._gtd = gtd;
	accountData._requests = [];
        accountData._loadForAccount(account);
        return accountData;
    },
    //}
    
    // Static Properties {
	
	
    /**
     * Dispatched when ready to load GTDI
     * @event ?
     */
    LOADED: 'loaded',
    
    /**
     * Dispatched when confirm/lose connection with GTDInbox server
     * @event ?
     */
    SYNC_CHANGED: 'sync_changed',
	
    //}
});
gtdAccountData.implement(glEventDispatcher); // Add event dispatching support
