/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * gtdRequest is an XMLHttpRequest wrapper for queries to GTDInbox servers. 
 *
 * @class ?
 */
var gtdAccountRequest = glBase.extend({
    // Properties {

    /**
     * The URL that the request will use
     * @property {String} ?
     */
    url: null,

    /**
     * The account used in the last connect attempt
     * @property {String} ?
     */
    account: null,

    /**
     * The password used in the last connect attempt
     * @property {String} ?
     */
    password: null,

    /**
     * Reference to the gtdMainPagePlugin for the page
     * @property {gtdMainPagePlugin}
     */
    _gtd: null,

    /**
     * The post data in a string
     * @property {String} ?
     */
    _postString: null,

    /**
     * The XMLHttpRequest object
     * @property {XMLHttpRequest} ?
     */
    _xmlRequest: null,
    //}

    /**
     * The constructor.
     * If one of the properties in params or postData is set to an array, that
     * property is added multiple times - once for each value in the array.
     * For the 'at' property, passing in a value of true will cause it to automatically
     * be calculated and added to the url params. The 'rt' property is optional
     * but if not set, it defaults to 'j' (which requires the least overhead to
     * process).
     * @constructor ?
     * @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
     * @param {String} path - The path for the URL
     * @param {Object} urlParams - The URL params.
     * @param {optional Object} postData - The post data
     */
    constructor: function(gtd, path, urlParams, postData) {
        if(typeof urlParams == "undefined") glLogger.error('urlParams is undefined', 'gtdAccountRequest');
        this._gtd = gtd;

        // Generate URL including url parameters {
        var url = gtdAccountRequest.GTDI_URL;
        url += path;
        url += glRequest._convertToRequestString(urlParams);
        this.url = url;
        //}

        // Generate POST string {
        if(typeof postData != "undefined") {
            this._postString = glRequest._convertToRequestString(postData);
        }
        
        
        glLogger.log('constructed gtdAccountRequest: ' + url, 'gtdAccountRequest.constructor');
        //}
    },

    // Methods {
    /**
     * Runs the request either synchronously or asynchronously
     * @function ?
     * @param {optional Boolean} async - Whether or not to make the call asynchronously, defaults to true
     */
    run: function(async) {
        glLogger.log('run request for: ' + this.url, 'gtdAccountRequest.run');
        if(async == undefined) async = true;
        this.reset();

        this._xmlRequest = new XMLHttpRequest();

        var me = this;
        this._xmlRequest.onload = function(evt) {me._requestCallback(evt)};
        this._xmlRequest.onerror = function(evt) {me._requestError(evt)}

        if(this._postString == null) {
            glLogger.log('Making '+(async?'asynchronous':'synchronous')+' GET XMLHttpRequest to "'+this.url+'"', 'gtdAccountRequest');
            this._xmlRequest.open("GET", this.url, async);
            this._xmlRequest.send(null);
        } else {
            glLogger.log('Making '+(async?'asynchronous':'synchronous')+' POST XMLHttpRequest to "'+this.url+'"'+"\n"+'Post data: "'+this._postString+'"', 'gtdAccountRequest');
            this._xmlRequest.open("POST", this.url, async);
            this._xmlRequest.setRequestHeader("Content-type", "application/x-www-form-urlencoded");
            this._xmlRequest.setRequestHeader("Content-length", this._postString.length); 
            this._xmlRequest.setRequestHeader("Connection", "close"); 
            this._xmlRequest.send(this._postString);
        }
        
    },

    /**
     * The request callback from the XMLHttpRequest, after reaching the server.
     * @function ?
     */
    _requestCallback: function(evt) {
        glLogger.log('request callback for: ' + this.url, 'gtdAccountRequest._requestCallback');
        glLogger.log("Response:\n"+this._xmlRequest.responseText, 'gtdAccountRequest._requestCallback');
        
        
        try{
            // Received data
            try{
                var evalJS = 'var result = (' + this._xmlRequest.responseText + ')';
                var sb = Components.utils.Sandbox(this.url);
                Components.utils.evalInSandbox(evalJS, sb);
            }catch(e){
                this.dispatchEvent({name: gtdAccountRequest.FAIL});
                this._xmlRequest = null;
                return;
            }
            if(typeof sb.result != 'object' ){
                this.dispatchEvent({name: gtdAccountRequest.FAIL});
                // There was a problem
            }else{
                // Received valid data
                if( sb.result.status=='ok' ){
                    this.dispatchEvent({name: gtdAccountRequest.SUCCESS, result: sb.result});
                }else if( sb.result.status=='wrong_password' ){
                    this.dispatchEvent({name: gtdAccountRequest.WRONG_PASSWORD});        
                }else if( sb.result.status=='no_account' ){
                    this.dispatchEvent({name: gtdAccountRequest.NO_ACCOUNT});
                }else{
                    // Unknown
                    this.dispatchEvent({name: gtdAccountRequest.FAIL});
                }
            }       
            
            this._xmlRequest = null;
        }catch(e){
            glLogger.warn(e, 'gtdAccountRequest._requestCallback');
            this.dispatchEvent({name: gtdAccountRequest.FAIL});
            this._xmlRequest = null;
        }
    },

    /**
     * The request callback from the XMLHttpRequest in case of error
     * @function ?
     */
    _requestError: function(evt) {
        glLogger.log('error for: ' + this.url, 'gtdAccountRequest._requestError');
        this.dispatchEvent({name: gtdAccountRequest.FAIL});
        this._xmlRequest = null;
    },

    /**
     * Cleans up the remnants of the previous run, preserving the URL and POST data
     * @function ?
     */
    reset: function() {
        glLogger.log('Reseting the gtdAccountRequest object', 'gtdAccountRequest');
        if(this._xmlRequest) this._xmlRequest.abort();
        this._xmlRequest = null;

    }
    //}
},{
    
    GTDI_URL: 'http://www.gtdinbox.com/',
    
    // Static Properties {
    /**
     * Dispatched when the request is successful
     * @event ?
     */
    SUCCESS: 'success',
    
    /**
     * Dispatched when the request failed (e.g. could not reach server)
     * @event ?
     */
    FAIL: 'fail',
    
    /**
     * Dispatched when the request failed due to a wrong password
     * @event ?
     */
    WRONG_PASSWORD: 'wrong_password',
    
    /**
     * Dispatched when the request failed due to a user not having an account
     * @event ?
     */
    NO_ACCOUNT: 'no_account',
    //}

    // Static Methods {
    

    /**
     * Creates and returns a request that adds the given label to the given
     * thread(s).
     * @function {static glRequest} ?
     * @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
     * @param {String} label - The label to add
     * @param {mixed} threads - The thread or threads to add the label to
     * @return The created request object
     */
    createHasAccountRequest: function(gtd, account) {
        return new gtdAccountRequest(
            gtd,
            'hasaccount',
            {},
            {}
        );
    },

    /**
     * Creates and returns a request that adds the given label to the given
     * thread(s).
     * @function {static glRequest} ?
     * @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
     * @param {String} label - The label to add
     * @param {mixed} threads - The thread or threads to add the label to
     * @return The created request object
     */
    createReadSettingsRequest: function(gtd, account) {
        return new gtdAccountRequest(
            gtd,
            'readsettings',
            {},
            {}
        );
    },
    
    
    /**
     * Creates and returns a request that adds the given label to the given
     * thread(s).
     * @function {static glRequest} ?
     * @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
     * @param {String} label - The label to add
     * @param {mixed} threads - The thread or threads to add the label to
     * @return The created request object
     */
    createWriteSettingsRequest: function(gtd) {
        return new gtdAccountRequest(
            gtd,
            'writesettings',
            {},
            {}
        );
    },
    

    //}
});
gtdAccountRequest.implement(glEventDispatcher); // Add event dispatching support
