/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * Store hierarchial label data. 
 * 
 * TODO Performance overhaul:
 * * Only build gtdLabelsTree on demand
 * * Teach .getLabelsArray et all to work purely off label data obtained from environment
 *
 * @class ? @extends glBase
 */
var gtdLabelsData = glBase.extend({
    	// Properties {

    	/** @property {private glShowHider[]} ? Array of ShowHiders in use  */
    	_showHiders: null,
    
    	/** @property {private gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page  */
    	_gtd: null,
    	
    	/** @property {private gtdLabelsBox} */
    	_labelsBox: null,
    	
    	/** @property {private Object} ? Cache of gtdLabelsTree elements */
    	_labelsTrees: null,
	
	/** @property {private Object} ? Collection of labels broken by type. Format  _labels[type] = String[] */
	_labels: null,
	
    	//}
	

	// Methods {
    
    
    	/**
    	* @constructor ?
    	* @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
    	*/
    	constructor: function(gtd){
	    	try{
	    		this.base(gtd.gmail);
	    		this._gtd = gtd;
	    		this._labelsTrees = {};
	    		this._labels = {};
	    		
	    		this._gtd.gmail.environment.addEventListener(glEnvironment.ENVIRONMENT_CHANGED, this._refreshLabels, this);
	    		this._gtd.gmail.environment.update();	// get latest labels
	    		this._refreshLabels(true);
	    		this._labelsBox = this._gtd.getUIObject('labelsBox');	// TODO test environment ok for this class
		}catch(e){
			glLogger.error('Error constructing gtdLabelsData\n' + e.toString(), 'gtdLabelsData');
		}
    	},
    	
    	/**
    	* Clean up
    	* @function ?
    	*/
    	unload: function(){
    		this.base();
    		this._gtd = null;

    		this._labelsBox.disconnect();
    	},
    	
    	/** 
    	* Clear out the gtdLabelsTree cache and rebuild them
    	* @function {} ?
    	*/
    	_refreshLabels: function(){
    		// Clear out old values:
    		var oldLabelsHash = {};
    		var firstLoad = true;
    		for( t in this._labels ){
    			if( firstLoad && this._labels[t].length>0 ) firstLoad = false;	// Has a label
    			oldLabelsHash[t] = glBase.arrayToHash(this._labels[t]);
    		}
		this._labels = {};
		for( t in this._labelsTrees ){
			this._labelsTrees[t].disconnect();
		}
		this._labelsTrees = {};
		
		// Prep:
		var typeSettings = this._gtd.accountData.getPref('labels.types');
		var prefixes = {};
		for( t in typeSettings ) if(t!='misc') prefixes[t] = new RegExp("^" + glBase.prepRE(typeSettings[t].prefix), "i");
		
		// TODO figure out if labels added/removed prior to any changes, then broadcast if so.
		// TODO ensure labelling box listens for these updates; and creates a new label structure if changes
		
		// Process labels
		var type = null;
		var added = [];
		var labels = this._gtd.gmail.environment.getUserLabels();
		for( var i = 0; i < labels.length; i++ ){
			if( !type || !prefixes[type] || !prefixes[type].test(labels[i].name) ){
				// Type has changed
				type = 'misc';
				for( t in typeSettings ){
					if( prefixes[t] && prefixes[t].test(labels[i].name) ){
						type = t;
						break;
					}
				}
			}
			if( !this._labels[type] ) this._labels[type] = [];
			this._labels[type].push(labels[i].name);
			
			if( oldLabelsHash[labels[i].name] ) delete oldLabelsHash[labels[i].name];
			else added.push(labels[i].name);
		}
		removed = glBase.hashKeysToArray(oldLabelsHash);
		
		if( !firstLoad ) if( added.length > 0 || removed.length > 0 ) this.dispatchEvent({name: gtdLabelsData.LABEL_CHANGED, added: added, removed: removed});
    	},
	
	/**
	* Get the type for a given label
	* @function {String} ?
	* @param {String} labelname - The labelname
	* @return {String} The type (e.g. statuses)
	*/
	getLabelType: function(labelname){	// TODO This is generic, should probably live in gtdLabelsBox; which 'pure' (non UI) parts should be abstracted out for
		var typeSettings = this._gtd.accountData.getPref('labels.types');
		var type = "misc";
		for( t in typeSettings ){
			if( gtdLabelsData.testPrefix(labelname, typeSettings[t].prefix) ){
				type = t;
				break;
			}
		}
		return t;
	},
	
	/**
    	* Clone a collection of labels (by type) within the typeBuckets data object (a collection of LabelData objects).
    	* @function ?
    	* @param {string} type - The type (collection of labels) to clone and return
    	* @return {Object} typeObject - Collection of gtdLabelDOM objects and container el for labels. Format: {el:, labels:{<labelname>:gtdLabelDOM}}
    	*/	
	getLabelsEl: function(type){
	
		var doc = this._gtd.gmail.canvasDocument;
		
		if( this._labelsTrees[type] ){
			return this._labelsTrees[type].clone();
		}else{
			this._labelsTrees[type] = this._gtd.getUIObject('labelsTree');
			this._labelsTrees[type].init(type);
			for( var i = 0; i < this._labels[type].length; i++ ) this._labelsTrees[type].addLabel( this._labels[type][i] );
			this._labelsTrees[type].finishAddingLabels();
			return this._labelsTrees[type];	// Note it is getting the original version
		}
		
	},
	
	
	/**
    	* Returns an array of labels for a given type (e.g. statuses, projects, etc.)
    	* @function ?
    	* @param {Array} types - The types to find the associated labels of. Can be '_all' (which is default if left blank), and returns all labels.
    	* @param {boolean} sortStatuses - If true and type is 'statuses', then returns core statuses first and in defined order.
    	* @param {boolean} removePlaceHolders - If true, ensures no place holder labels are returned (place holders are interims in hierarchy - e.g. for label 'p/gtdinbox/bugs', 'p/gtdinbox' is a place holder).
    	*/
	getLabelsArray: function(types, sortStatuses){
	
		if( !types ) types = ["_all"];
		else if( typeof types=="string" ){
			types = [types];
		}
		
		
		var labelsArray = [];
				
		if( types.length==1 ){
			if( types[0]=="_all" ){
				for( t in this._labels ){
					labelsArray = labelsArray.concat(this._labels[t]);
				}
			}else{
				if( sortStatuses && types[0]=="statuses" ){
					var statuses = this.getLabelsArray("statuses");
					var defaults = this._gtd.accountData.getPref('labels.types').statuses.core;
					var status = this._gtd.accountData.getStatusData(true);
					for( d in defaults ){	// Process the core in order
						for( var i = statuses.length - 1; i >=0; i-- ){	// Try to find a match in known statuses
							if( statuses[i]==status[d] ){	// It's a core status; add in order and remove from future consideration
								labelsArray.push(statuses[i]);
								statuses.splice(i, 1);
								break;
							}
						}
					}
					for( var i = 0; i < statuses.length; i++ ) labelsArray.push(statuses[i]);	// Add what statuses remain
				}else{
					if( this._labels[types[0]] ){
						labelsArray = labelsArray.concat(this._labels[types[0]]);
					}
				}
			}
		}else{
			for( var i = 0; i < types.length; i++ ){
				labelsArray = labelsArray.concat(this._labels[types[i]]);
			}
		}
		
		
		return labelsArray;
	
	},
	
	
	/**
    	* Returns a hash of labels for a given type (e.g. statuses, projects, etc.)
    	* @function ?
    	* @param {Array} types - The types to find the associated labels of. Can be '_all' (which is default if left blank), and returns all labels.
    	* @param {boolean} removePlaceHolders - If true, ensures no place holder labels are returned (place holders are interims in hierarchy - e.g. for label 'p/gtdinbox/bugs', 'p/gtdinbox' is a place holder).
    	*/
	getLabelsHash: function(types){
		if( !types ) types = ["_all"];
		else if( typeof types=="string" ) types = [types];
		
		var labelsHash = {};
				
		if( types.length==1 ){
			if( types[0]=="_all" ){
				for( t in this._labels ){
					glBase.arrayToHash( this._labels[t], false, labelsHash );
				}
			}else{
				if( this._labels[types[0]] ){
					glBase.arrayToHash( this._labels[types[0]], false, labelsHash );
				}
			}
		}else{
			for( var i = 0; i < types.length; i++ ){
				glBase.arrayToHash( this._labels[types[i]], false, labelsHash );
			}
		}
		return labelsHash;
	},
	
	
	/**
    	* Return true with the tested labelname exists
    	* @function ?
    	* @param {string} labelname - The labelname to test the existence of
    	* @param {boolean} removePlaceHolders - If true, do not allow tests for place holder names to pass
    	*/
	hasLabel: function(labelname, removePlaceHolders){
		// TODO This is used quite a lot; could cache the labels found into a hash for faster recall; or use prefix to 

		for( t in this._labels ){
			if( this._labels[t].indexOf(labelname)>-1 ) return true;
		}
		return false;
	},
	
	

    
    
},{
    // Static Properties {

    /**
     * Dispatched when new label detected (will not fire on first run)
     * @event ?
     */
    LABEL_CHANGED: 'label_changed',
    

    //}
    
    
	testPrefix: function(labelname, prefix){
		return (new RegExp("^" + glBase.prepRE(prefix), "i")).test(labelname);
	},
	
	removePrefix: function(labelname, prefix){
		return labelname.replace( new RegExp("^" + glBase.prepRE(prefix), "i"), "" );
	},
    
});

gtdLabelsData.implement(glEventDispatcher); // Add event dispatching support