/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * GTD plugin for the main Gmail page.
 *
 * @class ?
 */
var gtdMainPagePlugin = glBase.extend({
    // Properties {
    /**
     * glGmail object
     * @property {glGmail} ?
     */
    gmail: null,

    /**
     * Account data object for the current page
     * @property {gtdAccountData} ?
     */
    accountData: null,

    /**
     * The GTD Pre-labeler object for the current page
     * @property {gtdPreLabeler} ?
     */
    preLabeler: null,
    
    /**
    * The GTD Labelbox object for the current page. Refreshes itself everytime the label box does; and holds the hierarchy of labels.
    */
    labelsBox: null,

	/** @property {gtdEnvironment} ? Listen to clicks */
	environment: null,
	
	/** @property {gtdThreadContainer} */
	statusThreadsContainer: null,
	
	/** @property {gtdThreadRules} */
	threadRules: null,
	
	/** @property {gtdRecent} */
	recent: null,

    /**
     * A hash of all the UI objects
     * @property {Object} ?
     */
    _ui: null,
    //}

    /**
     * The constructor.
     * Initializes gtdAccountData object in preparation for being asked for
     * required glGmail components.
     * @constructor ?
     * @param {String} account - The user's account for the current page
     */
    constructor: function(account) {
        glLogger.log('gtdMainPagePlugin object created', 'gtdMainPagePlugin');

	this.accountData = gtdAccountData.retrieveForAccount(this, account);

        // Initialize UI object container
        this._ui = {};
    },

    // Methods {
    /**
     * Called by plugin architecture when plugins are ready to load (by main.js)
     * This isn't our primary loader - that is _load, called by accountData after reading settings.
     * @function ?
     * @param {glGmail} gmail - The glGmail object for the page
     */
    load: function(gmail) {
        glLogger.log('Plugin loading', 'gtdMainPagePlugin');
        this.gmail = gmail;

	// Load the international properties for the entire gtd plugin
        glI18N.addStringBundle("gtd");

        // Load account data object using given account	
	this.accountData.addEventListener(gtdAccountData.LOADED, this._load, this);
	this.accountData.load();
	
        
        
        
        //this.gmail.page.addEventListener(glPageWrapper.VIEW_CHANGED, this._viewChanged, this);
        
        // Wait for UI_READY event to finish loading
        //this.gmail.page.addEventListener(glPageWrapper.UI_READY, this._finishLoading, this);
    },
    
    /**
     * Called when gtdAccountData has loaded (including server interaction or install).
     */
    _load: function(){
	if( this.gmail.page.isUIReady() ){
	    this._finishLoading();
	    this._viewChanged();
	}else{
	    // Wait for UI_READY event to finish loading
	    this.gmail.page.addEventListener(glPageWrapper.UI_READY, this._finishLoading, this);  
	}
	// Listen for view changed:
	this.gmail.page.addEventListener(glPageWrapper.VIEW_CHANGED, this._viewChanged, this);
    },

    /**
     * Called on page unload, it cleans up everything
     * @function ?
     */
    unload: function() {
	if( this.statusThreadsContainer ){
		this.statusThreadsContainer.preventUnload = false;
		this.statusThreadsContainer.unload();
	}
		
        this.accountData.shutdown();
        
        this.labelsData.unload();
	this.environment.unload();
		
        for(var element in this._ui) {
            this._ui[element].disconnect();
        }
    },

    /**
     * Returns a list of the required glGmail components
     * @function {Array} ?
     * @return The required glGmail components
     */
    getRequiredComponents: function() {
        var requiredComponents = [];
        //if(this.accountData.getPref('components.traffic_watcher.enabled'))	// TODO Finalise removal. No reason to be optional (GTD relies too heavily on it); and confuses the situation now we don't guarantee to have settings loaded. Alternative solutions include activating/deactivating the traffic watcher post-LOAD; or having a post-LOAD mechanism to add components (in the plugin architecture).
	requiredComponents.push('trafficWatcher');

        return requiredComponents;
    },

    /**
     * Initializes a gtdSearch object and returns it
     * @function {gtdSearch} ?
     * @return The searcher object
     */
    getSearcher: function() {
        return new gtdSearch(this);
    },

    /**
     * Returns the given UI object after initializing it.
     * For example, if you wanted a gtdMainView object, you would call
     * getUIObject('mainView'), dropping the gtd and turning the 'M' to an
     * 'm'. If the object is cacheable and has already been loaded, it returns
     * the one in the cache.
     * @function {gtdUIObject} ?
     * @param {String} type - The UI object to return
     * @return The initialized UI object
     */
    getUIObject: function(type) {
        glLogger.log('Getting '+type+'...', 'gtdMainPagePlugin');

        var object = null;
        if(!this._ui[type]) {
            // Not in cache so create and cache (or return if not supposed to cache)

            // Add gtd to the begining of the type and raise the first character up
            // mainView -> gtdMainView
            theClass = 'gtd'+type.substr(0,1).toUpperCase()+type.substr(1);

            glLogger.log('Creating object: "'+theClass+'"', 'glGmail');

            try {
                object = new window[theClass]();
            } catch(e) {
                // Use glLogger.error because it gives a stack trace
                glLogger.error("Failure instantiating "+theClass+" ui object:\n"+e.message, 'gtdMainPagePlugin');
            }
	
            	object.connect(this);
		
            // Is the ui object cachable - return if not
            if(object.isCacheable()) {
                // Cache the instantiated object
                this._ui[type] = object;
            }
        } else {
            // Get object from cache
            object = this._ui[type];
        }

        return object;
    },

	/**
	* Get a gtdThreadContainer loaded with statuses
	* @function {gtdThreadContainer} ?
	*/	
	createStatusThreadsContainer: function(){
		// Create the query string:
		var status = this.accountData.getStatusData();
		var statuses = this.labelsData.getLabelsArray("statuses", false, true);
		var orConditions = new glSearchOrConditions();
		for( var i = 0; i < statuses.length; i++){
			if( statuses[i]!=status.finished ){
				orConditions.addCondition(glSearch.l(statuses[i]));
			}
		}
		var search = new glSearch(this.gmail);
		search.addCondition(orConditions);
		search.addCondition('-'+glSearch.l(status.finished));
		
		glLogger.log('query string\n' + search.getQuery(), 'gtdMainPagePlugin.getStatusThreadContainer');
		
		// Create the gtdThreadContainer with this query string:
		if( !this.statusThreadsContainer ){
	    		this.statusThreadsContainer = gtdThreadContainer.createWithSearch(this, search);
			this.statusThreadsContainer.addEventListener(gtdThreadContainer.CHANGED, function(event){
				glLogger.log("\n\nStatus Thread Container change", "gtdMainPagePlugin.getStatusThreadsContainer");
			}, this);
	    		this.statusThreadsContainer.search();
			this.statusThreadsContainer.preventUnload = true;
		}else{
			this.statusThreadsContainer.update(search);
		}
		

		return this.statusThreadsContainer;
	},

    /**
     * Finishes all the loading that GTDGmail needs. A callback to the
     * {@link glPageWrapper.UI_READY} event.
     * @function ?
     * @param {Event} evt - The UI_READY event
     */
    _finishLoading: function(evt) {
			// Remove event listener 
			this.gmail.page.removeEventListener(glPageWrapper.UI_READY, this._finishLoading);

			this.gmail.fb = this.gmail.canvasDocument.defaultView.wrappedJSObject.top.console;

			// Add the CSS // TODO make part of plugin structure
			glGmail.addCSS(this.gmail.canvasDocument, "chrome://gtdinbox/skin/glGmail.css");
			glGmail.addCSS(this.gmail.canvasDocument, "chrome://gtdinbox/skin/gtd/c/reset.css");
			glGmail.addCSS(this.gmail.canvasDocument, "chrome://gtdinbox/skin/gtd/c/main.css");

			// Load the Labels box: 
			this.labelsData = new gtdLabelsData(this); 
			
			// Load the environment box:
			this.environment = new gtdEnvironment(this);
			
			// Load the thread rules
			this.threadRules = new gtdThreadRules(this);
			
			// Load the recent
			this.recent = new gtdRecent(this);
			
			// TODO either in the gtdThreadContainer, or below, add listeners that refresh the statusThreadContainer when labels change.

			// Load the threadContainer for statuses
			this.createStatusThreadsContainer();
			this.labelsData.addEventListener(gtdLabelsData.LABEL_CHANGED, glBase.bind(this, function(event){
				// Label changed. 
				// If add a label, it can be added to a thread upon creation. If that is a status label, it may mean we need to add a new thread.
				// If rename a label, it means our search query can be out of date.
				// If we remove a label, we may no longer wish to include threads that had that label.
				// So, in all cases, we need to update the query to use latest statuses; then update.
				this.createStatusThreadsContainer();
			}), this);

			// Load gtdMainView, which loads all the rest of the GTD widgets
			var mainView = this.getUIObject('mainView');
		

    },

    /**
     * Triggered by the {@link glPageWrapper.VIEW_CHANGED} event, it calls
     * getCurrentViewUIObject to make sure it gets connected. This is the wrong
     * behavior, but necessary for now.
     * @param {Event} evt - The VIEW_CHANGED event
     */
    _viewChanged: function(evt) {
    	try{
	    	glLogger.log("start view changed", "gtdMainPagePlugin");
	        
	        var view = this.gmail.getCurrentViewUIObject();	// Ensure it gets connected
		   
		
	        // In the cache, so disconnect
	        if(this._ui['currentViewObject']){
			this._ui['currentViewObject'].disconnect();
			this._ui['currentViewObject'] = null;
		}
	
	
	        // Get the UI object type for a call to getUIObject
	        var uiObject = '';
	        switch(this.gmail.page.getActiveViewType()) {
	            case 'tl': uiObject = 'threadListView'; break;
	            case 'cv': uiObject = 'conversationView'; break;
	            case 'co': uiObject = 'composeView'; break;
	        }
	
	        if(uiObject) {
	            // Use getUIObject so extending classes can override the view objects without overriding this method
	            this._ui['currentViewObject'] = this.getUIObject(uiObject);
	        }
	        
	
	    	glLogger.log("end view changed", "gtdMainPagePlugin");
	}catch(e){
		glLogger.error(glLogger.dump(e, true), 'gtdMainPagePlugin._viewChanged', this.gmail);
	}
    },
    
    
    /**
     * Returns the UI object for the current view
     * @function {glUIView} ?
     * @return The view UI object
     */
    getCurrentViewUIObject: function() {
        // In the cache, so return
        if(typeof this._ui['currentViewObject'] != 'undefined') {
            return this._ui['currentViewObject'];
        }
    },
    
    test: function() {}
    //}
});


// Register plugin
if(typeof window['glPlugins'] == 'undefined') {
    if(typeof window['glPluginsList'] == 'undefined') var glPluginsList = [];
    glPluginsList.push(['gtdMainPagePlugin', 'FULL_PAGE']);
} else {
    glPlugins.register('gtdMainPagePlugin', 'FULL_PAGE');
}
