/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * The label add/remover UI for any thread object. 
 *
 * @class ? @extends gtdUIObject
 */
var gtdLabellingBox = gtdUIObject.extend({
    // Properties {
    
	/** @property {private gtdMainPagePlugin} ? The main gtd plugin for this page */
	_gtd: null,
	
	/** @property {private boolean} ? If true, when a label is clicked, it is automatically added/removed from the thread. */
	_autoUpdate: null,
	
	/** @property {Object} ? A hash of label rows. Format: _uiRows[<type>] = {el:, labelEls:{<labelName>:el}, labellingDropdown:gtdLabellingDropdown, disabled:true/false} */
	_uiRows: null,
	
    	/** @property {private glEventsContainer} ? Collection of manageable events on DOM elements */
	_eventsContainer: null,
	
	/** @property {glThread} ? The associated thread object */
	thread: null,
	
	/** @property {Element} ? The label container */
	el: null,
	
	/** @property {Element} ? The row container (child of 'el') */
	elRows: null,
	
	/** @property {private String} ? Method for updating server (e.g. labels) */
	_serverActionType: null,
	
	/** @property {Element} ? Contains archive/finish/etc. */
	_elCmdButtons: null,
	
	/** @property {Element} ?  */	
	_elArchive: null,
	
	/** @property {Element} ?  */	
	_elFinish: null,
	
    //}

	// Methods {
	/**
	* Connects the element to the given gtdMainPagePlugin object's page. Called
	* automatically by {@link gtdMainPagePlugin.getUIObject}.
	* @function ?
	* @param {gtdMainPagePlugin} gtd - The gtdMainPagePlugin object for the page
	*/
	connect: function(gtd) {
		glLogger.log('Connecting gtdLabellingBox object to view', 'gtdLabellingBox');
	
		this.base(gtd);
		this._gtd = gtd;
		this._eventsContainer = new glEventsContainer(this);
		this._uiRows = {};
		
		// Create the nodes:
		var doc = gtd.gmail.canvasDocument;
		this.el = doc.createElement("DIV");
			this.el.className = "gtdi";
		
		var elInner = doc.createElement("DIV");
			elInner.className = "gtdi-labelling";
		this.el.appendChild( elInner );
		this.elRows = doc.createElement("UL");
		elInner.appendChild( this.elRows );
		
		
		// Create basic structure:
		this._createLabelStructure();
		
		// Listen for changes that will require a structural change:
		this._gtd.labelsData.addEventListener(gtdLabelsData.LABEL_CHANGED, this._createLabelStructure, this);
		
		// Listen for changes to threads
		this._gtd.gmail.actions.addEventListener(glActionManager.THREAD_CHANGED, this._threadUpdated, this);
		
		// Listen for clicks
		this._eventsContainer.observe( this.el, "click", this._click, false );

	},
	

    /**
     * Called on page shutdown, it does cleanup
     * @function ?
     */
    disconnect: function() {
        glLogger.log('Disconnecting gtdLabellingBox from view', 'gtdLabellingBox');
        
        for( t in this._uiRows ){
	        if(this._uiRows[t].labellingDropdown) this._uiRows[t].labellingDropdown.disconnect();
		}
		
		this._gtd.labelsData.removeEventListener(gtdLabelsData.LABEL_CHANGED, this._createLabelStructure, this);
		this._gtd.gmail.actions.removeEventListener(glActionManager.THREAD_CHANGED, this._threadUpdated, this);
        
        if( this.el.parentNode ) this.el.parentNode.removeChild( this.el );
        this.el = null;
        this.elRows = null;
        
        this._eventsContainer.reset();
        this._eventsContainer = null;
        
        this.base();
    },
	
	/**
	* Clear the container el, and create a row for each label type, and a row for the expander
	* If thread loaded, reload it
	*/
	_createLabelStructure: function(){
		// Clear out any existing structure:
		for( r in this._uiRows ){
			if( this._uiRows[r].labellingDropdown ) this._uiRows[r].labellingDropdown.disconnect();
		}
		this.elRows.innerHTML = "";
		this._uiRows = {};
		
		
		// Retrieve the types
		var doc = this._gtd.gmail.canvasDocument;
		var typeSettings = this._gtd.accountData.getPref('labels.types');
		labelsData = this._gtd.labelsData;
		
		
		// Add a row for each type
		var hasHiddenRows = false;
		for( t in typeSettings ){
			if( labelsData.getLabelsArray(t).length>0 ){
				
				this._uiRows[t] = {};
				this._uiRows[t].labelEls = {};
				this._uiRows[t].disabled = (labelsData.getLabelsArray(t).length==0 )? true : false;
				
				this._uiRows[t].el = doc.createElement("LI");
				if( t=="statuses" ){
					this._uiRows[t].el.className = "gtdi-row-actions";
					var elUL = doc.createElement("UL");
					this._uiRows[t].el.appendChild( elUL );
					
					
					var statuses = labelsData.getLabelsArray("statuses", true, true);
					var statusData = this._gtd.accountData.getStatusData(true);
					for( var i = 0; i < statuses.length; i++ ){
						var key = statusData.reverseLookup[statuses[i]];
						if( key!="finished" ){
							if( key=="today" ){
								var elLi = doc.createElement("LI");
								elLi.className = "gtdi-label-status gtdi-label-"+key;
								elLi.setAttribute("gtdLabelName", statuses[i]);
								elLi.innerHTML = "<a href='#'>" + glI18N.getString("Labelling.statuses."+key, statuses[i]) + "</a>";
								elUL.appendChild( elLi );
							}else{
								var elLi = doc.createElement("LI");
								elLi.className = "gtdi-label-status gtdi-label-" + key;
								elLi.setAttribute("gtdLabelName", statuses[i]);
								elLi.innerHTML = "<a href='#'><b><span class='gtdi-labelname'>" + glI18N.getString("Labelling.statuses." + key, statuses[i]) + "</span></b></a>";
								elUL.appendChild( elLi );
							}
						}
					}
					
					
					// Add the Archive/Finish buttons
					var elButtons = doc.createElement("DIV");
					elButtons.id = "gtdi-arcfin";
					elButtons.className = "gtdi-btngrp";
					this._uiRows[t].el.appendChild( elButtons );
					
					this._elCmdButtons = doc.createElement("DIV");
					this._elCmdButtons.className = "gtdi-btn-grp";
					elButtons.appendChild( this._elCmdButtons );
					
					this._elArchive = doc.createElement("DIV");
					this._elArchive.className = "gtdi-btn";
					this._elArchive.innerHTML = "<b class='gtdi-btn-t'></b><b class='gtdi-btn-t2'></b><b class='gtdi-btn-m'></b><b class='gtdi-btn-b2'></b><b class='gtdi-btn-b'></b><span class='gtdi-btn-label'><span class='gtdi-archive'>" + glI18N.getString("Labelling.buttons.archive") + "</span></span>";
					this._elCmdButtons.appendChild( this._elArchive );
					
					this._elFinish = doc.createElement("DIV");
					this._elFinish.className = "gtdi-btn";
					this._elFinish.innerHTML = "<b class='gtdi-btn-t'></b><b class='gtdi-btn-t2'></b><b class='gtdi-btn-m'></b><b class='gtdi-btn-b2'></b><b class='gtdi-btn-b'></b><span class='gtdi-btn-label'><span class='gtdi-finish'>" + glI18N.getString("Labelling.buttons.finish") + "</span></span>";
					this._elCmdButtons.appendChild( this._elFinish );					
					
				}else{
					
					this._uiRows[t].el.className = "gtdi-row";
					this._uiRows[t].el.innerHTML = "<h3>" + glI18N.getString("Labels.Types." + t, t) + "</h3>";
					this._uiRows[t].labellingDropdown = this._gtd.getUIObject("labellingDropdown");
					this._uiRows[t].labellingDropdown.setType(t);
					this._uiRows[t].el.appendChild( this._uiRows[t].labellingDropdown.el );
					
				}
				
				if( this._uiRows[t].disabled ) this._uiRows[t].el.style.display = "none";
				if( !(t=="statuses" || t=="projects") ){
					hasHiddenRows = true;
					this._uiRows[t].el.style.display = "none";	// leave to expander, which always starts closed
				}
				
				this.elRows.appendChild( this._uiRows[t].el );
			}
		}
		
		
		if( hasHiddenRows ){
			// Add the expander row:
			var elExpander = doc.createElement("LI");
			elExpander.className = "gtdi-labelling-expander";
			elExpander.innerHTML = "<a href='#expandcontract'><img alt='" + glI18N.getString("Labelling.expander.opener") + "' src='chrome://gtdinbox/skin/gtd/i/gtdi-dropdown-arrow.png'/></a>";
			elExpander.title = glI18N.getString("Labelling.expander.opener");
			this.elRows.appendChild( elExpander );
		}
		this._testCmdButtons();
	},
	
	/**
	* Associate a thread with this box, load its labels and listen for changes to that thread.
	* @function ?
	* @param {glThread} - The thread object
	* @param {Boolean} autoUpdate - If true, when a label is clicked, it is automatically added/removed from the thread.
	*/
	load: function(thread, autoUpdate){
		// Disassociated any existing thread
		if( this.thread ){
			this.thread = null;
		}
		
		// Attach the new thread
		this.thread = thread;
		
		this._autoUpdate = autoUpdate;
				
		// Select labels
		if( this.thread ){
			this.uiSelectLabels( this.thread.labels );
			
			// Set cmds
			this._testCmdButtons();
			this._elCmdButtons.style.display = "";
		}else{
			this._elCmdButtons.style.display = "none";
		}
		
	},
	
	/**
	* Update the UI to show all provided labels as 'selected' and all other labels as unselected.
	* Used by pre-labelling in Compose; and when loading a thread
	* @function ?
	* @param {string[]} - Array of label names to mark as selected
	*/
	uiSelectLabels: function(labels){
		// Mark all previous labels as unselected (including dropdowns):
		
		var doc = this._gtd.gmail.canvasDocument;
		
		// Get active labels:
		var currentLabels = {};
		var result = doc.evaluate(".//*[contains(@class,'gtdi-active')]", this.elRows, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
		for( var i = 0; i < result.snapshotLength; i++ ){
			currentLabels[ result.snapshotItem(i).getAttribute("gtdLabelName") ] = true;
		}
		// Remove any still desired labels, or any private ones:
		for( var i = labels.length - 1; i>=0; i-- ){
			if( currentLabels[labels[i]] ) delete currentLabels[labels[i]];
		}
		// Turn active labels off:
		for( l in currentLabels ){
			this._setLabel(l, false);
		}
		
		
		// Mark labels as selected:
		for( var i = 0; i < labels.length; i++ ){
			if( labels[i].indexOf("^")==-1 ) this._setLabel( labels[i], true );
		}
	},
	
	
	/**
	* Visually set a label as selected or unselected in the box, and propagate to dropdowns.
	* @function ?
	* @param {string} labelname - The name of the label to update
	* @param {boolean} isOn - If true, mark selected. Otherwise, mark unselected.
	*/
	_setLabel: function(labelname, isOn){
		if( !labelname ) return;
		// Find the type
		var foundType = null;
		var typeSettings = this._gtd.accountData.getPref('labels.types');
		for( t in typeSettings ){
			if( gtdLabelsData.testPrefix(labelname, typeSettings[t].prefix) ){
				foundType = t;
				break;
			}
		}
		if( !foundType || !this._uiRows[foundType] ) return;
		
		
		// Choose how to handle the label. Statuses are treated uniquely.
		var doc = this._gtd.gmail.canvasDocument;
		if( foundType=="statuses" ){
			// Get the relevant label
			var elLabel = null;
			if( labelname.indexOf("'")>-1 ){	// Contains ', do long way. TODO - find native XPath solution for apostrophe
				var result = doc.evaluate("./li[@gtdlabelname]", this._uiRows[foundType].el, null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);
				for( var i = 0; i < result.snapshotLength; i++ ){
					if( result.snapshotItem(i).getAttribute("gtdLabelName")==labelname ){
						elLabel = result.snapshotItem(i);
						break;
					}
				}
			}else{
				elLabel = doc.evaluate(".//li[@gtdlabelname='" + labelname + "']", this._uiRows[foundType].el, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
			}
			if( elLabel ){
				if( isOn ) glUIWidget.addClassName( elLabel, "gtdi-active" );	// Mark it selected
				else glUIWidget.removeClassName( elLabel, "gtdi-active" );	// Mark it unselected
			}
		}else{
			if( isOn ){	// Wants to be on
				if( !this._uiRows[foundType].labelEls[labelname] ){	// Does not have label element
					// Add label element
					var elLabel = doc.createElement("DIV");
					elLabel.className = "gtdi-label gtdi-active";
					elLabel.setAttribute("gtdLabelName", labelname);
					elLabel.innerHTML = "<span class='gtdi-labelname'>" + labelname + "</span><span class='gtdi-labelremove'>x</span>";
					this._uiRows[foundType].el.appendChild( elLabel );
					this._uiRows[foundType].labelEls[labelname] = elLabel;
				}
			}else{	// Wants to be off
				if( this._uiRows[foundType].labelEls[labelname] ){	// Has label element
					// Remove label element
					this._uiRows[foundType].labelEls[labelname].parentNode.removeChild( this._uiRows[foundType].labelEls[labelname] );
					this._uiRows[foundType].labelEls[labelname] = null;
				}
			}
			// Now notify the labellingDropdown
			this._uiRows[foundType].labellingDropdown.setLabel(labelname, isOn);
		}
	},
	
	
	/**
	* Set method for updating server (e.g. labels)
	* @function ?
	* @param {String} type - The type. A static variable of glActionManager
	*/
	setServerActionType: function(type){
		this._serverActionType = type;
	},
	
	
	/**
	* Respond to a click event.
	* Could be on a status button, on a label 'X', or the expander.
	* @function ?
	*/
	_click: function(event){
		try{
			var el = event.target;
			if( el.className=="gtdi-labelremove" ){	// Clicked 'X'
				var labelname = el.parentNode.getAttribute("gtdLabelName");
				this._removeLabel(labelname);
			}else if( el.getAttribute('gtdLabelName') ){
				this._addLabel( el.getAttribute('gtdLabelName') );
			}else if( el.className=="gtdi-archive" ){
				if( this.thread.hasLabel("^i") ){
					var serverActionType = this._gtd.gmail.page.getActiveViewType()=='cv'? glActionManager.USE_AJAX : this._serverActionType;
					this._gtd.gmail.actions.performAction(glActionManager.MARK_AS_READ, this.thread.getRecentId(), serverActionType);
					this._gtd.gmail.actions.performAction(glActionManager.ARCHIVE, this.thread.getRecentId(), serverActionType);
				}
			}else if( el.className=="gtdi-finish" ){
				var statusData = this._gtd.accountData.getStatusData();
				if( !this.thread.hasLabel(statusData.finished) ){
					this._addLabel( statusData.finished );
					var serverActionType = this._gtd.gmail.page.getActiveViewType()=='cv'? glActionManager.USE_AJAX : this._serverActionType;
					this._gtd.gmail.actions.performAction(glActionManager.MARK_AS_READ, this.thread.getRecentId(), serverActionType);
					this._gtd.gmail.actions.performAction(glActionManager.ARCHIVE, this.thread.getRecentId(), serverActionType);	// TODO Review use of IDs. It seems archiving is very fussy about last ID, whereas normal labels are not. 
				}
			}else{					// Work up tree to understand it
				while( el.parentNode ){
					if( el.className.indexOf("gtdi-labelling-expander")>-1 ){	// Clicked expander - toggle
						var elImg = el.getElementsByTagName("IMG")[0];
						if( elImg.src=="chrome://gtdinbox/skin/gtd/i/gtdi-dropdown-arrow.png" ){	// Currently closed, open it
							for( t in this._uiRows ){
								if( !(t=="statuses" || t=="projects" ) ) this._uiRows[t].el.style.display = "";
							}
							elImg.alt = glI18N.getString("Labelling.expander.closer");
							el.title = glI18N.getString("Labelling.expander.closer");
							elImg.src = "chrome://gtdinbox/skin/gtd/i/gtdi-up-arrow.png";
						}else{	// Currently open, close it
							for( t in this._uiRows ){
								if( !(t=="statuses" || t=="projects" ) ) this._uiRows[t].el.style.display = "none";
							}
							elImg.alt = glI18N.getString("Labelling.expander.opener");
							el.title = glI18N.getString("Labelling.expander.opener");
							elImg.src = "chrome://gtdinbox/skin/gtd/i/gtdi-dropdown-arrow.png";
						}
						break;
					}else if( el.className.indexOf("gtdi-label-status")>-1 ){	// Clicked status label - toggle
						// Stop link being followed:
						glBase.stopEvent(event);
						// Toggle:
						var labelname = el.getAttribute("gtdLabelName");
						if( glUIWidget.hasClassName( el, "gtdi-active" ) ){
							this._removeLabel(labelname);
						}else{
							this._addLabel(labelname);
						}
						break;
					}
					// Go up tree
					el = el.parentNode;
				}
			}
		}catch(e){
			glLogger.warn(e, 'gtdLabellingBox._click', this._gtd.gmail);
		}
	},
	
	/**
	* Respond to a request to add a label (could originate from _click or a broadcast event from gtdLabellingDropdown)
	* If autoUpdate, update associated thread. Broadcast event to any listeners.
	* @function ?
	* @param {} event - An event object (or acceptably, a string) containing the label name.
	*/
	_addLabel: function(event){
		// Get the label name
		var labelname = (typeof event=="string")? event : event.label;
		// Process it?
		
		if( this.thread && this._autoUpdate ){
			this._gtd.gmail.actions.performAction( glActionManager.ADD_LABEL, this.thread, this._serverActionType, labelname );
		}
		
		// Set it as on - the thread should call this again in <1000ms when it confirms it as added. This is just for a fast user response (or if manual).
		this._setLabel(labelname, true);
				
		// Broadcast
		if( this.hasEventListener(gtdLabellingBox.SELECT_LABEL)) this.dispatchEvent({name: gtdLabellingBox.SELECT_LABEL, label:labelname});
	},
	
	/**
	* Respond to a request to remove a label (could originate from _click or a broadcast event from gtdLabellingDropdown)
	* If autoUpdate, update associated thread. Broadcast event to any listeners.
	* @function ?
	* @param {} event - An event object (or acceptably, a string) containing the label name.
	*/	
	_removeLabel: function(event){
		// Get the label name
		var labelname = (typeof event=="string")? event : event.label;
		
		// Process it?
		if( this.thread && this._autoUpdate ){
			this._gtd.gmail.actions.performAction( glActionManager.REMOVE_LABEL, this.thread, this._serverActionType, labelname );
		}
		
		// Set it as off - the thread should call this again in <1000ms when it confirms it as added. This is just for a fast user response (or if manual).
		this._setLabel(labelname, false);
				
		// Broadcast
		if( this.hasEventListener(gtdLabellingBox.UNSELECT_LABEL)) this.dispatchEvent({name: gtdLabellingBox.UNSELECT_LABEL, label:labelname});
	},
	
	/**
	* Respond to a thread having a label added/removed. Update the UI of the box to show the new label selected (and pass to dropdowns)
	* @function ?
	* @param {} event - Contains the labelname and the action {action: const, thread: id}
	*/
	_threadUpdated: function(event){
			var statusData = this._gtd.accountData.getStatusData();
			if( this.thread && this.thread.hasId(event.thread) ){	// It affects us
				if( (event.label && event.label==statusData.finished) || (event.action==glActionManager.ARCHIVE) ){
					this._testCmdButtons();
				}
				if( event.action==glActionManager.ADD_LABEL ){		// Label added to thread
					this._setLabel( event.label, true );
				}else if( event.action==glActionManager.REMOVE_LABEL ){	// Label removed from thread
					this._setLabel( event.label, false );
				}
			}
	},

	/**
	* Retrieve the container for the command buttons (Archive/Finish)
	* @function {Element} ?
	*/
	getCmdButtons: function(){
		return this._elCmdButtons;
	},
	
	/**
	* Set UI of test buttons based on environment
	* @function {} ?
	*/
	_testCmdButtons: function(){
		if( !this.thread ) return;
		var statusData = this._gtd.accountData.getStatusData();
		
		var elArchiveInner = this._elArchive.ownerDocument.evaluate(".//*[@class='gtdi-archive']", this._elArchive, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
		var elFinishInner = this._elFinish.ownerDocument.evaluate(".//*[@class='gtdi-finish']", this._elFinish, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;			
		
		if( this.thread.hasLabel("^i") ){	// Not archived
			glUIWidget.removeClassName(this._elArchive, "gtdi-active");
			this._elArchive.style.color = "#090";
			this._elArchive.style.fontWeight = "bold";
			elArchiveInner.innerHTML = glI18N.getString("Labelling.buttons.archive");
		}else{
			glUIWidget.addClassName(this._elArchive, "gtdi-active");
			this._elArchive.style.color = "";
			this._elArchive.style.fontWeight = "normal";
			elArchiveInner.innerHTML = glI18N.getString("Labelling.buttons.archived");
			
			if( !this.thread.hasLabel(statusData.finished) ){
				this._elFinish.style.color = "#090";
				this._elFinish.style.fontWeight = "bold";
			}
		}
		
		if( this.thread.hasLabel(statusData.finished) ){
			glUIWidget.addClassName(this._elFinish, "gtdi-active");
			this._elFinish.style.color = "";
			this._elFinish.style.fontWeight = "normal";
			elFinishInner.innerHTML = glI18N.getString("Labelling.buttons.finished");
		}else{
			glUIWidget.removeClassName(this._elFinish, "gtdi-active");
			elFinishInner.innerHTML = glI18N.getString("Labelling.buttons.finish");
		}
	},
    
    /**
     * Whether the UI object can be cached by {@link gtdMainPagePlugin.getUIObject}
     * @function {Boolean} ?
     * @return Whether the UI can be cached
     */
    isCacheable: function() {
        return false;
    },
    
    //}
},{
    // Static Properties {
    /**
     * Dispatched when a label is selected
     * @event ?
     */
    SELECT_LABEL: 'select_label',
    
    /**
     * Dispatched when a label is unselected
     * @event ?
     */
    UNSELECT_LABEL: 'unselect_label',

});
gtdLabellingBox.implement(glEventDispatcher); // Add event dispatching support