/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Andy Mitchell (andy@gtdinbox.com)
 * 
 * ***** END LICENSE BLOCK ***** */



/**
 * Provides internal tracing and debugging functionality. It supports object
 * dumping, fast message logging, error/stack trace logging, and it can be easilly
 * disabled for release builds.
 *
 * @class ?
 */
var glLogger = glBase.extend(null, {
    // Static Properties {
    /**
     * Whether or not logging is enabled
     * @property {static} ?
     */
    disabled: true,

    /**
     * The list of filter rules to apply to logged messages.
     * Filtering is applied against the optional context in the {@link log} or
     * {@link error} calls. Put a minus then a space to make it a negative rule,
     * and a plus-sign for a postive rule. After the space goes the context to act
     * on, or '*' for everything. By default, all contexts are included.<br />
     * <br />
     * For example:<br />
     * <code>filterRules: ['- *', '+ gtdAccountData', '+ glDataObject']</code>
     * @property {static String[]} ?
     */
    filterRules: [],

    /**
     * Mozilla service for logging
     * @property {static nsIConsoleService} ?
     */
    _consoleService: Components.classes['@mozilla.org/consoleservice;1'].getService(Components.interfaces.nsIConsoleService),
    //}

    // Static Methods {
    /**
     * Logs the given message to the console
     * @function {static} ?
     * @param {String} msg - The message to log
     * @param {optional String} context - The context of where the message originated
     */
    log: function(msg, context) {
        if(glLogger.disabled) return;

        var logMsg = "";
        if(msg && msg.constructor == String) {
            if(context != undefined) {
                if(glLogger._notFiltered(context)) {
                    logMsg = context+': '+msg;
                } else {
                    return;
                }
            } else {
                logMsg = msg;
            }
        } else {
            logMsg = "Log: nothing";
        }

        glLogger._output(logMsg);
    },

    /**
     * Logs the passed in message as an error and throws an exception
     * @function {static} ?
     * @param {String} msg - The message to log
     * @param {optional String} context - The context of where the error originated
     */
    error: function(msg, context) {
        if(glLogger.disabled) return;

        glLogger.warn(msg, context);

        throw new Error(msg.toString()); // After error message logged, throw an exception
    },
    
    /**
     * Logs the passed in message as an error
     * @function {static} ?
     * @param {String} msg - The message to log
     * @param {optional String} context - The context of where the error originated
     * @param {optional glGmail} glGmail - If provided, write the error to the Firebug console (if available)
     */
    warn: function(msg, context, glGmail) {
        if(glLogger.disabled) return;

        var errorMsg = "";
        if(context != undefined) {
            errorMsg = "Error in "+context+": "+msg.toString();
        } else {
            errorMsg = "Error: "+msg.toString();
        }

        try {
            throw new Error();
        } catch(e) {
            var stack = e.stack.toString().split("\n");
            var stackMessage = "Exception thrown\n"+stack.slice(2).join("\n");

            errorMsg += "\n\nStack trace:\n"+stackMessage;
        }

        glLogger._output(errorMsg);
        if( glGmail && glGmail.canvasDocument && glGmail.canvasDocument.defaultView.wrappedJSObject.top.console && glGmail.canvasDocument.defaultView.wrappedJSObject.top.console.log )glGmail.canvasDocument.defaultView.wrappedJSObject.top.console.log(errorMsg);
    },

    /**
     * Converts the given object into a printable string and logs it.
     * If doReturn is true, then it returns the string instead of logging it
     * @function {static String} ?
     * @param {Object} obj - The object to dump
     * @param {optional Boolean} doReturn - Whether or not to return the created string or simply log it
     * @param {optional int} level - The indentation level (FOR INTERNAL USE ONLY)
     * @return The logged string if doReturn is true
     */
    dump: function(obj, doReturn, level) {
        if(glLogger.disabled) return null;
        if(doReturn == undefined) doReturn = false;
        if(level == undefined) level = 0;

        var returnStr = "";
        var indent = "";
        for(var l = 0; l < level; l++) {
            indent += "\t";
        }

        if(obj === null) {
            returnStr = 'null';
        } else if(obj === undefined) {
            returnStr = 'undefined';
        } else if(obj.constructor.name == 'String') {
            returnStr = '"'+obj+'"';
        } else if(typeof obj == 'number') {
            returnStr = obj.toString();
        } else if(typeof obj == 'boolean') {
            returnStr = (obj)?'true':'false';
        } else if(obj.constructor.name == 'Array') {
            var arrayStr = "";
            if(obj.length) {
                arrayStr += "[\n";
                for(var i = 0; i < obj.length; i++) {
                    arrayStr += indent+"\t"+glLogger.dump(obj[i], false, level+1)+",\n";
                }
                arrayStr += indent + "]";
            } else {
                arrayStr += "[]";
            }
            returnStr = arrayStr;
        } else if(obj.constructor.name == 'Object') {
            var objStr = "{\n";
            var foundProps = false;
            for(var prop in obj) {
                foundProps = true;
                objStr += indent+"\t"+prop+": "+glLogger.dump(obj[prop], false, level+1)+",\n";
            }
            objStr += indent + "}";
            if(!foundProps) {
                objStr = "{}";
            }
            returnStr = objStr;
        } else if(obj && obj.toString) {
            returnStr = obj.toString();
        } else if(obj && obj.constructor) {
            returnStr = obj.constructor.toString();
        } else {
            returnStr = 'invalid';
        }

        if(level == 0) {
            if(doReturn) {
                return returnStr;
            } else {
                glLogger._output(returnStr);
                return null;
            }
        } else {
            return returnStr;
        }
    },

    /**
     * Method for logging final strings to the console
     * @function {static} ?
     * @param {String} msg - The message to output
     */
    _output: function(msg) {
        glLogger._consoleService.logStringMessage(msg);
    },

    /**
     * Checks whether or not a given context is filtered
     * @function {static Boolean} ?
     * @param {String} context - The context to check
     * @return Whether or not it's filtered
     */
    _notFiltered: function(context) {
        var filteredOut = false;
        for(var i = 0; i < glLogger.filterRules.length; i++) {
            var rule = {
                type: glLogger.filterRules[i].slice(0,1),
                rule: glLogger.filterRules[i].slice(2)
            };

            if(rule['type'] == '-') {
                if(rule['rule'] == '*' || rule['rule'] == context) {
                    filteredOut = true;
                }
            } else if(rule['type'] == '+') {
                if(rule['rule'] == '*' || rule['rule'] == context) {
                    filteredOut = false;
                }
            } else {
                continue;
            }
        }

        return !filteredOut;
    }
    //}
});
