/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Andy Mitchell (andy@gtdinbox.com)
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * ***** END LICENSE BLOCK ***** */


/**
 * A UI helper that makes showing/hiding content boxes easy.
 *
 * @interface glShowHider
 */
 
var glShowHider = glBase.extend({

	// Properties {
	
	/**
     	* The glGmail object for the page
    	* @property {private glGmail} ?
     	*/
	_gmail: null,
	
	/**
	* The gtdMainPagePlugin object for the page
	* @property {private gtdMainPagePlugin} ?
	*/
	_gtd: null,
    
    	/** @property {private glEventsContainer} ? Collection of manageable events on DOM elements */
	_eventsContainer: null,
	
	/**
     	* The ID for this Show/Hider. Used to store show/hide status long-term
    	* @property {int} ?
     	*/
	id: null,
	
	/**
     	* The container for the entire ShowHider
    	* @property {Element} ?
     	*/
	el: null,
	
	/**
     	* The container for the title elements in the ShowHider (elTriangle and elTitleTxt)
    	* @property {Element} ?
     	*/
	elTitle: null,
	
	/**
     	* The image used to show whether the ShowHider is shown or hidden
    	* @property {Element} ?
     	*/
	elTriangle: null,
	
	/**
     	* The text that displays the title of this ShowHider
    	* @property {Element} ?
     	*/
	elTitleTxt: null,
	
	/**
     	* The content element
    	* @property {Element} ?
     	*/
	elContent: null,


	// }
	
	// Methods {
	
	
	/**
    	* The constructor.
	* Initializes the object.
    	* @constructor ?
     	* @param {glGmail} gmail - The glGmail object for the page
     	* @param {gtdMainPagePlugin} gtd - Optional.
     	* @param {string} id - The unique ID for this ShowHider. Used to remember show/hide status long-term
     	* @param {Object} options - A collection of options for how the ShowHider is created. 
     	* @param {boolean} show - Force the starting show/hide status of the ShowHider
	*/
	constructor: function(gmail, gtd, id, options, show) {
	        glLogger.log('glShowHider object created -> ' + id, 'glShowHider');
	
		this._gmail = gmail;
		this._gtd = gtd;
		this._eventsContainer = new glEventsContainer(this);
		this.id = id;
				
		if( options ) this.create(options);
		
		// Decide show status:
		if( typeof show!="undefined" ){
			this.show(show);
		}else if( this._gtd ){	// Read preferences
			var showHiders = this._gtd.accountData.getPref('showHiders', true) || {};
			if( typeof showHiders[this.id]!='undefined' ) this.show(showHiders[this.id]);
			else this.show(true);
		}
	},

	/**
    	* Create the element structure
    	* @function ?
     	* @param {Object} options - A collection of options for how the ShowHider is created. 
	*/
	create: function(options){
		// Create the ShowHider:
		var doc = this._gmail.canvasDocument;
		if( !options.type ) options.type = "create";

		this.elTitleTxt = doc.createElement("SPAN");
		this.elTriangle = doc.createElement("IMG");
		switch( options.type ){
			case "create":	// Create a new element structure
				this.el = doc.createElement("DIV");
					this.el.className = "showhider";
	
				if( options.elParent && options.elBefore ) options.elParent.insertBefore( this.el, options.elBefore );
				else if( options.elParent ) options.elParent.appendChild( this.el );
				
				this.setTitle(options.title);
				this.elTitle = doc.createElement("DIV");
				this.elContent = options.elContent;
					if( !this.elContent ) this.elContent = doc.createElement("DIV");
				this.el.appendChild( this.elTitle );
				this.el.appendChild( this.elContent );
				break;
			case "attach":	// Merge into an existing element structure
				this.el = options.el;
				var els = this.el.getElementsByTagName("*");
				for( var i = 0; i < els.length; i++ ){
					if( glUIWidget.hasClassName(els[i], options.titleClass) ){
						this.elTitle = els[i];
						break;
					}
				}
				this.setTitle( this.elTitle.innerHTML.replace(/<.+?>/g,"") );
				this.elTitle.innerHTML = "";
				this.elContent = this.elTitle.nextSibling;
				if( !this.elContent ){
					this.elContent = doc.createElement("DIV");
					this.el.appendChild( this.elContent );
				}
				break;
		}
		this.elTitle.appendChild( this.elTriangle );
		this.elTitle.appendChild( this.elTitleTxt );
			
		glUIWidget.addClassName(this.elTitle, "showhider-title");
		glUIWidget.addClassName(this.elContent, "showhider-content");
		
		this.elTitle.style.cursor = "pointer";
		if( !options.inline ){
			this.elTitleTxt.style.marginLeft = "3px;"
			if( !options.noContentIndent ) this.elContent.style.paddingLeft = "14px";
		}
		this._eventsContainer.observe( this.elTitle, "mousedown", this._mouseDownTitle, false );
		
		var elClear = doc.createElement("DIV");
			elClear.style.clear = "both";
			elClear.style.height = "1px";
		this.elTitle.appendChild( elClear );
	},

	/**
	* Update the text of the title element
	* @function ?
	* @param {string} t - The new title
	*/
	setTitle: function(t){
		this.elTitleTxt.innerHTML = t;
	},

	/**
	* Set the content to either visible or invisible; or if no parameter supplied, return the show status.
	* @function ?
	* @param {boolean} v - True: make it shown; false: make it hidden; undefined: return true/false
	*/
	show: function(v){
		if( typeof v=="undefined" ) return this.elContent.style.display=="";
		
		
		if( v ){
			this.elTriangle.src = "chrome://gtdinbox/skin/opentriangle.gif";
			glUIWidget.addClassName(this.el, "showhider-open");
			this.elContent.style.display = "";
			if(this.hasEventListener(glShowHider.SHOW)) this.dispatchEvent({name: glShowHider.SHOW});
			if(this.hasEventListener(glShowHider.TOGGLE)) this.dispatchEvent({name: glShowHider.TOGGLE});
		}else{
			glUIWidget.removeClassName(this.el, "showhider-open");
			this.elTriangle.src = "chrome://gtdinbox/skin/triangle.gif";
			this.elContent.style.display = "none";
			if(this.hasEventListener(glShowHider.HIDE)) this.dispatchEvent({name: glShowHider.HIDE});
			if(this.hasEventListener(glShowHider.TOGGLE)) this.dispatchEvent({name: glShowHider.TOGGLE});
		}
	},

	/**
	* Event in response to a mousedown. Toggle the show/hide status.
	* @function ?
	* @param {Event} event - 
	*/	
	_mouseDownTitle: function(event){
		this.toggle();
		
		if( this._gtd ){
			var showHiders = this._gtd.accountData.getPref('showHiders', true) || {};
			showHiders[this.id] = this.show();
			this._gtd.accountData.setPref('showHiders', showHiders);
			this._gtd.accountData.savePrefs();
		}
	},

	/**
	* Invert the show/hide status
    * @function ?
	*/
	toggle: function() {
		this.show( !this.show() );
    },
   
   	/**
   	* Set the entire ShowHider as visible or not.
   	* @function ?
   	* @param {boolean} v - If true, set visible. If false, set hidden. If undefined, return visible status.
   	*/
	visible: function(v){
		if( typeof v=="undefined" ) return (this.el.style.display!="none");
		this.el.style.display = v? "" : "none";
	},
    	
    	/**
    	* Releases items from memory
    	* @function ?
	*/
    	reset: function(){
    		this._eventsContainer.reset();
    		this.id = null;
    		this.el = null;
    		this.elTitle = null;
    		this.elTriangle = null;
    		this.elTitleTxt = null;
    		this.elContent = null;
    	},
    	//}


},{
    // Static Properties {

    /**
     * Dispatched when content made visible
     * @event ?
     */
    SHOW: 'show',

    /**
     * Dispatched when content hidden
     * @event ?
     */
    HIDE: 'hide',
    
    /**
     * Dispatched when ever Show status is modified
     * @event ?
     */
    TOGGLE: 'toggle',
    
    //}
});
glShowHider.implement(glEventDispatcher); // Add event dispatching support