/*
 * ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is GTDInbox.
 *
 * The Initial Developer of the Original Code is
 * Stephen Augenstein
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * ***** END LICENSE BLOCK ***** */



/**
 * Stores a list of registered plugins and provides accessors for getting them.
 *
 * <h3>Plugin Interface:</h3>
 * Plugins must conform to the interface below, or they will not be loaded.
 * <pre>
 * /*
 *  * Constructor
 *  * Called with the email address for the current page's user. This can be used
 *  * to dynamically adjust the result of the getRequiredComponents method
 *  * /
 * var CLASS_NAME = function(accountEMailAddress) {}
 *
 * /*
 *  * Called when ready to load
 *  * Parameters
 *  *   {glGmail} gmail - The glGmail object for the page
 *  * /
 * CLASS_NAME.prototype.load = function(glGmailObject) {}
 *
 * /*
 *  * Called on page unload, it cleans up everything
 *  * /
 * CLASS_NAME.prototype.unload = function() {}
 *
 * /*
 *  * Returns a list of the required glGmail components for the plugin
 *  * Returns
 *  *   {Array} - The required glGmail components
 *  * /
 * CLASS_NAME.prototype.getRequiredComponents = function() {}
 * </pre>
 *
 * <h3>Plugin Registration:</h3>
 * To register your plugin, simply replace CLASS_NAME with the name of the plugin
 * class in the following code:
 * <pre>
 * // Register plugin
 * if(typeof window['glPlugins'] == 'undefined') {
 *     if(typeof window['glPluginsList'] == 'undefined') var glPluginsList = [];
 *     glPluginsList.push(['CLASS_NAME', 'FULL_PAGE']);
 * } else {
 *     glPlugins.register('CLASS_NAME', 'FULL_PAGE');
 * }
 * </pre>
 *
 * @class ?
 */
var glPlugins = glBase.extend(null, {
    // Static Properties {
    /**
     * Hash of all registered plugins indexed by page type
     * @property {static Object} ?
     */
    _plugins: {},

    /**
     * A reference to the nsIJSON interface
     * @property {static nsIJSON} ?
     */
    JSON: Components.classes["@mozilla.org/dom/json;1"].createInstance(Components.interfaces.nsIJSON),
    //}

    // Static Methods {
    /**
     * Registers all plugins that were loaded before glPlugins was.
     * @function {static} ?
     */
    initialize: function() {
        if(typeof window['glPluginsList'] == 'undefined' || !glPluginsList.length) return;

        for(var i = 0; i < glPluginsList.length; i++) {
            glPlugins.register.apply(null, glPluginsList[i]);
        }
        glPluginsList = undefined;
    },

    /**
     * Registers the given plugin for page loads of the given type
     * @function {static} ?
     * @param {String} pluginClass - The class of the plugin as a string
     * @param {String} pageType - The page type to instantiate the plugin for, like a full page, the print page, etc.
     */
    register: function(pluginClass, pageType) {
        if(typeof glPlugins._plugins[pageType] == 'undefined') glPlugins._plugins[pageType] = [];
        glLogger.log('Registering plugin "'+pluginClass+'" for '+pageType, 'glPlugins');
        glPlugins._plugins[pageType].push(pluginClass);
    },

    /**
     * Returns all registered and active plugins for the given page type and
     * account. Plugin activation settings are stored as a JSON string in
     * "extensions.gtdinbox.disabledPlugins". If a plugin with the name of
     * "glTestPlugin" is not expressly disabled and is registered, then only that
     * plugin will be returned by this method.
     * @function {static String[]} ?
     * @param {String} pageType - The page type to instantiate the plugin for, like a full page, the print page, etc.
     * @param {String} account - The gmail account for the current page (if available)
     * @return The registered and active plugin classes
     */
    getRegisteredFor: function(pageType, account) {
        var disabledPlugins = {};
        try {
            var prefManager = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
            var prefString = prefManager.getCharPref("extensions.gtdinbox.disabledPlugins");
            var disabled = glPlugins.JSON.decode(prefString);
            if(typeof disabled[account] != 'undefined') {
                disabledPlugins = disabled[account];
            }
        } catch(e) {}

        // Disable glTestPlugin unless explicit setting
        if(typeof disabledPlugins['glTestPlugin'] == 'undefined') disabledPlugins['glTestPlugin'] = true;

        var possiblePlugins = (typeof glPlugins._plugins[pageType] == 'undefined') ? [] : glPlugins._plugins[pageType];
        var enabledPlugins = [];
        for(var i = 0; i < possiblePlugins.length; i++) {
            if(!disabledPlugins[possiblePlugins[i]]) {
                if(possiblePlugins[i] == 'glTestPlugin') {
                    // Test plugin is enabled, so don't enable any other registered plugins
                    return ['glTestPlugin'];
                } else {
                    enabledPlugins.push(possiblePlugins[i]);
                }
            }
        }

        return enabledPlugins;
    },
    //}
});
glPlugins.initialize();